import pathlib

import numpy as np
import pytest
from astropy.nddata import StdDevUncertainty
from astropy.wcs import WCS
from ndcube import NDCube

from punchbowl.data import NormalizedMetadata
from punchbowl.level1.deficient_pixel import remove_deficient_pixels

THIS_DIRECTORY = pathlib.Path(__file__).parent.resolve()


@pytest.fixture()
def sample_bad_pixel_map(shape: tuple = (2048, 2048), n_bad_pixels: int = 20) -> NDCube:
    """
    Generate some random data for testing
    """
    bad_pixel_map = np.ones(shape)

    x_coords = np.fix(np.random.random(n_bad_pixels) * shape[0]).astype(int)
    y_coords = np.fix(np.random.random(n_bad_pixels) * shape[1]).astype(int)

    bad_pixel_map[x_coords, y_coords] = 0
    bad_pixel_map = bad_pixel_map.astype(int)

    uncertainty = StdDevUncertainty(bad_pixel_map)

    wcs = WCS(naxis=2)
    wcs.wcs.ctype = "HPLN-AZP", "HPLT-AZP"
    wcs.wcs.cunit = "deg", "deg"
    wcs.wcs.cdelt = 0.02, 0.02
    wcs.wcs.crpix = 1024, 1024
    wcs.wcs.crval = 0, 24.75
    wcs.array_shape = shape

    meta = NormalizedMetadata({"TYPECODE": "CL", "LEVEL": "1", "OBSRVTRY": "0", "DATE-OBS": "2008-01-03 08:57:00"})
    return NDCube(data=bad_pixel_map, uncertainty=uncertainty, wcs=wcs, meta=meta)


@pytest.fixture()
def perfect_pixel_map(shape: tuple = (2048, 2048)) -> NDCube:
    """
    Generate some random data for testing
    """
    bad_pixel_map = np.ones(shape)

    bad_pixel_map = bad_pixel_map.astype(int)

    uncertainty = StdDevUncertainty(bad_pixel_map)

    wcs = WCS(naxis=2)
    wcs.wcs.ctype = "HPLN-AZP", "HPLT-AZP"
    wcs.wcs.cunit = "deg", "deg"
    wcs.wcs.cdelt = 0.02, 0.02
    wcs.wcs.crpix = 1024, 1024
    wcs.wcs.crval = 0, 24.75
    wcs.array_shape = shape

    meta = NormalizedMetadata({"TYPECODE": "CL", "LEVEL": "1", "OBSRVTRY": "0", "DATE-OBS": "2008-01-03 08:57:00"})
    return NDCube(data=bad_pixel_map, uncertainty=uncertainty, wcs=wcs, meta=meta)


@pytest.fixture()
def one_bad_pixel_map(shape: tuple = (2048, 2048)) -> NDCube:
    """
    Generate pixel map with one bad pixel at 100, 100
    """
    bad_pixel_map = np.ones(shape)

    bad_pixel_map = bad_pixel_map.astype(int)

    bad_pixel_map[100, 100] = 0

    uncertainty = StdDevUncertainty(bad_pixel_map)

    wcs = WCS(naxis=2)
    wcs.wcs.ctype = "HPLN-AZP", "HPLT-AZP"
    wcs.wcs.cunit = "deg", "deg"
    wcs.wcs.cdelt = 0.02, 0.02
    wcs.wcs.crpix = 1024, 1024
    wcs.wcs.crval = 0, 24.75
    wcs.array_shape = shape

    meta = NormalizedMetadata({"TYPECODE": "CL", "LEVEL": "1", "OBSRVTRY": "0", "DATE-OBS": "2008-01-03 08:57:00"})
    return NDCube(data=bad_pixel_map, uncertainty=uncertainty, wcs=wcs, meta=meta)


@pytest.fixture()
def nine_bad_pixel_map(shape: tuple = (2048, 2048)) -> NDCube:
    """
    Generate pixel map with one bad pixel at 100, 100
    """
    bad_pixel_map = np.ones(shape)

    bad_pixel_map = bad_pixel_map.astype(int)

    bad_pixel_map[100,100] = 0
    bad_pixel_map[100,101] = 0
    bad_pixel_map[100,102] = 0
    bad_pixel_map[101,100] = 0
    bad_pixel_map[101,101] = 0
    bad_pixel_map[101,102] = 0
    bad_pixel_map[102,100] = 0
    bad_pixel_map[102,101] = 0
    bad_pixel_map[102,102] = 0

    uncertainty = StdDevUncertainty(bad_pixel_map)

    wcs = WCS(naxis=2)
    wcs.wcs.ctype = "HPLN-AZP", "HPLT-AZP"
    wcs.wcs.cunit = "deg", "deg"
    wcs.wcs.cdelt = 0.02, 0.02
    wcs.wcs.crpix = 1024, 1024
    wcs.wcs.crval = 0, 24.75
    wcs.array_shape = shape

    meta = NormalizedMetadata({"TYPECODE": "CL", "LEVEL": "1", "OBSRVTRY": "0", "DATE-OBS": "2008-01-03 08:57:00"})
    return NDCube(data=bad_pixel_map, uncertainty=uncertainty, wcs=wcs, meta=meta)


@pytest.fixture()
def increasing_pixel_data(shape: tuple = (2048, 2048)) -> NDCube:
    """
    Generate data of increasing values for testing; data[0,100]=0.0, data[100,0]=100.0
    """
    data = np.ones(shape)
    for iStep in range(2048):
        data[iStep, :] = iStep

    uncertainty = StdDevUncertainty(np.zeros_like(data))

    wcs = WCS(naxis=2)
    wcs.wcs.ctype = "HPLN-AZP", "HPLT-AZP"
    wcs.wcs.cunit = "deg", "deg"
    wcs.wcs.cdelt = 0.02, 0.02
    wcs.wcs.crpix = 1024, 1024
    wcs.wcs.crval = 0, 24.75
    wcs.array_shape = shape

    meta = NormalizedMetadata({"TYPECODE": "CL", "LEVEL": "1", "OBSRVTRY": "0", "DATE-OBS": "2008-01-03 08:57:00"})
    return NDCube(data=data, uncertainty=uncertainty, wcs=wcs, meta=meta)


@pytest.fixture()
def sample_punchdata(shape: tuple = (2048, 2048)) -> NDCube:
    """
    Generate a sample PUNCH data object for testing
    """

    data = np.random.random(shape)
    uncertainty = StdDevUncertainty(np.zeros_like(data))

    wcs = WCS(naxis=2)
    wcs.wcs.ctype = "HPLN-AZP", "HPLT-AZP"
    wcs.wcs.cunit = "deg", "deg"
    wcs.wcs.cdelt = 0.02, 0.02
    wcs.wcs.crpix = 1024, 1024
    wcs.wcs.crval = 0, 24.75
    wcs.array_shape = shape

    meta = NormalizedMetadata({"TYPECODE": "CL", "LEVEL": "1", "OBSRVTRY": "0", "DATE-OBS": "2008-01-03 08:57:00"})
    return NDCube(data=data, uncertainty=uncertainty, wcs=wcs, meta=meta)


def test_remove_deficient_pixels(sample_punchdata: NDCube, sample_bad_pixel_map: NDCube) -> None:
    """
    Test the remove_deficient_pixels prefect flow using a test harness, providing a filename
    """
    flagged_punchdata = remove_deficient_pixels(sample_punchdata, sample_bad_pixel_map.data)
    assert isinstance(flagged_punchdata, NDCube)
    assert np.all(np.isposinf(flagged_punchdata.uncertainty[np.where(sample_bad_pixel_map.data == 0)].array))


def test_nan_input(sample_punchdata: NDCube, sample_bad_pixel_map: NDCube) -> None:
    """
    The module output is tested when NaN data points are included in the input PUNCH data object. Test for no errors.
    """

    input_data = sample_punchdata
    input_data.data[42, 42] = np.nan

    flagged_punchdata = remove_deficient_pixels(input_data, sample_bad_pixel_map.data)

    assert isinstance(flagged_punchdata, NDCube)
    assert np.all(np.isposinf(flagged_punchdata.uncertainty[np.where(sample_bad_pixel_map.data == 0)].array))


def test_data_loading(sample_punchdata: NDCube, perfect_pixel_map: NDCube) -> None:
    """
    A specific observation is provided. The module loads it as a PUNCH data object.
    No bad data points, in same as out. uncertainty should be the same in and out.
    """
    deficient_punchdata = remove_deficient_pixels(sample_punchdata, perfect_pixel_map.data)

    assert isinstance(deficient_punchdata, NDCube)
    assert np.all(deficient_punchdata.data == sample_punchdata.data)
    assert np.all(deficient_punchdata.uncertainty.array == sample_punchdata.uncertainty.array)


def test_artificial_pixel_map(sample_punchdata: NDCube, sample_bad_pixel_map: NDCube) -> None:
    """
    A known artificial bad pixel map is ingested. The output flags are tested against the input map.
    """

    flagged_punchdata = remove_deficient_pixels(sample_punchdata, sample_bad_pixel_map.data)
    assert isinstance(flagged_punchdata, NDCube)
    assert np.all(np.isposinf(flagged_punchdata.uncertainty[np.where(sample_bad_pixel_map.data == 0)].array))


def test_data_window_1(increasing_pixel_data: NDCube, one_bad_pixel_map: NDCube) -> None:
    """
    dataset of increasing values passed in, a bad pixel map is passed in
    """
    deficient_punchdata = remove_deficient_pixels(increasing_pixel_data, one_bad_pixel_map.data)
    assert isinstance(deficient_punchdata, NDCube)
    assert (deficient_punchdata.data[5, 5] == increasing_pixel_data.data[5, 5])
    assert (deficient_punchdata.uncertainty.array[5, 5] == increasing_pixel_data.uncertainty.array[5, 5])
    assert (deficient_punchdata.data[100, 100] == 100)


def test_mean_data_window_1(increasing_pixel_data: NDCube, one_bad_pixel_map: NDCube) -> None:
    """
    dataset of increasing values passed in, a bad pixel map is passed in
    """
    deficient_punchdata = remove_deficient_pixels(increasing_pixel_data, one_bad_pixel_map.data, method='mean')

    assert isinstance(deficient_punchdata, NDCube)
    assert (deficient_punchdata.data[5, 5] == increasing_pixel_data.data[5, 5])
    assert (deficient_punchdata.uncertainty.array[5, 5] == increasing_pixel_data.uncertainty.array[5, 5])
    assert (deficient_punchdata.data[100, 100] == 100)


def test_data_window_9(increasing_pixel_data: NDCube, nine_bad_pixel_map: NDCube) -> None:
    """
    dataset of increasing values passed in, a bad pixel map is passed in
    """
    deficient_punchdata = remove_deficient_pixels(increasing_pixel_data, nine_bad_pixel_map.data)

    assert isinstance(deficient_punchdata, NDCube)
    assert (deficient_punchdata.data[5, 5] == increasing_pixel_data.data[5, 5])
    assert (deficient_punchdata.uncertainty.array[5, 5] == increasing_pixel_data.uncertainty.array[5, 5])
    assert (deficient_punchdata.data[101, 101] == 101)


def test_mean_data_window_9(increasing_pixel_data: NDCube, nine_bad_pixel_map: NDCube) -> None:
    """
    dataset of increasing values passed in, a bad pixel map is passed in
    """
    deficient_punchdata = remove_deficient_pixels(increasing_pixel_data, nine_bad_pixel_map.data, method='mean')

    assert isinstance(deficient_punchdata, NDCube)
    assert (deficient_punchdata.data[5, 5] == increasing_pixel_data.data[5, 5])
    assert (deficient_punchdata.uncertainty.array[5, 5] == increasing_pixel_data.uncertainty.array[5, 5])
    assert (deficient_punchdata.data[101, 101] == 101)



"""
===================
Environment setup
===================

To run the sample notebooks contained within punchbowl, you must first create a Python environment.
"""

# %%
# To run the sample notebooks contained within punchbowl, you must first create a Python environment. These days, it's recommended to have a local virtual environment for a project rather than a global configuration.
# If using an environment for running Jupyter notebooks (Visual Studio Code, PyCharm, etc), this environment setup may be handled automatically or through UI prompts on notebook execution.
# Note that for a full environment setup, including full PUNCH requirements, see the instructions linked at the `punch-mission README <https://github.com/punch-mission/punch-mission?tab=readme-ov-file#setting-up-the-punch-environment>`_.

# %%
# To set up manually, start within a terminal, and navigate to the directory in which you want to work with notebooks.
# Begin by setting up a virtual environment within the specified directory. Here we'll setup a hidden directory .venv
#
# .. code-block:: shell
#
#   python -m venv .venv
#

# %%
# Next activate the virtual environment by executing the activation script. Note that when returning to this environment in future, you'll also need to activate the environment once again.
#
# .. code-block:: shell
#
#   source .venv/bin/activate
#

# %%
# To get started quickly with punchbowl, one can install a few packages using pip. Other packages of note can be installed in this same way.
#
# .. code-block:: shell
#
#   pip install ipykernel jupyter punchbowl
#

# %%
# From here, one can open a downloaded notebook in an editor of choice (selecting the local virtual environment), or by calling jupyter directly.
#
# .. code-block:: shell
#
#   jupyter downloaded_notebook.ipynb
#



import astropy.wcs.utils
import numpy as np
import reproject
from astropy.wcs import WCS
from ndcube import NDCube

from punchbowl.data.wcs import calculate_celestial_wcs_from_helio
from punchbowl.prefect import punch_flow, punch_task


@punch_task
def reproject_cube(input_cube: NDCube, output_wcs: WCS, output_shape: tuple[int, int]) -> np.ndarray:
    """
    Core reprojection function.

    Core reprojection function of the PUNCH mosaic generation module.
        With an input data array and corresponding WCS object, the function
        performs a reprojection into the output WCS object system, along with
        a specified pixel size for the output array. This utilizes the adaptive
        reprojection routine implemented in the reprojection astropy package.

    Parameters
    ----------
    input_cube: NDCube
        input cube to be reprojected
    output_wcs
        astropy WCS object describing the coordinate system to transform to
    output_shape
        pixel shape of the reprojected output array

    Returns
    -------
    np.ndarray
        output array after reprojection of the input array


    Example Call
    ------------
    >>> reprojected_arrays = reproject_cube(input_cube, output_wcs, output_shape)

    """
    input_wcs = input_cube.wcs
    time = input_cube.meta.astropy_time

    celestial_source = calculate_celestial_wcs_from_helio(input_wcs, time, output_shape)
    celestial_target = calculate_celestial_wcs_from_helio(output_wcs, time, output_shape)

    # When we build mosaics, each input image fills only a portion (less than half) of the output frame. When we
    # reproject, we don't want it spending time looping over all those empty pixels, calculating coordinates,
    # etc. So here we find a bounding box around the input in the output frame and crop to that before reprojecting.
    # To start, here we make a grid of points along the edges of the input image.
    xs = np.linspace(-1, input_cube.data.shape[-1], 60)
    ys = np.linspace(-1, input_cube.data.shape[-2], 60)
    edgex = np.concatenate((xs, # bottom edge
                            np.full(len(ys), xs[-1]), # right edge
                            xs, # top edge
                            np.full(len(ys), xs[0]))) # left edge
    edgey = np.concatenate((np.full(len(xs),ys[0]), # bottom edge
                            ys, # right edge
                            np.full(len(xs), ys[-1]), # top edge
                            ys)) # left edge

    # Now we transform them to the output frame
    xs, ys = astropy.wcs.utils.pixel_to_pixel(celestial_source, celestial_target, edgex, edgey)
    # And we find that bounding box
    xmin, xmax = int(np.floor(xs.min())), int(np.ceil(xs.max()))
    ymin, ymax = int(np.floor(ys.min())), int(np.ceil(ys.max()))
    xmin = np.max((xmin, 0))
    ymin = np.max((ymin, 0))
    xmax = np.min((xmax, output_shape[1]))
    ymax = np.min((ymax, output_shape[0]))

    output_array = np.full((2, *output_shape), np.nan)
    input_data = np.stack([input_cube.data, input_cube.uncertainty.array])
    # Reproject will complain if the input and output arrays have different dtypes
    input_data = np.asarray(input_data, dtype=float)

    reproject.reproject_adaptive(
        (input_data, celestial_source),
        celestial_target[ymin:ymax, xmin:xmax],
        shape_out=(2, np.max((ymax-ymin, 0)), np.max((xmax-xmin, 0))),
        roundtrip_coords=False, return_footprint=False,
        output_array=output_array[..., ymin:ymax, xmin:xmax],
    )

    return output_array


@punch_flow
def reproject_many_flow(data: list[NDCube | None], trefoil_wcs: WCS, trefoil_shape: np.ndarray) -> list[NDCube | None]:
    """Reproject many flow."""
    out_layers = [reproject_cube.submit(d, trefoil_wcs, trefoil_shape) if d is not None else None for d in data]

    return [NDCube(data=out_layers[i].result()[0],
                   uncertainty=out_layers[i].result()[1],
                   wcs=trefoil_wcs,
                   meta=d.meta) if d is not None else None for i, d in enumerate(data)]



import os.path
import warnings
from math import ceil

import numpy as np
from ndcube import NDCube

from punchbowl.exceptions import DataValueWarning
from punchbowl.prefect import punch_task

TABLE_PATH = os.path.dirname(__file__) + "/decoding_tables/"


def decode_sqrt(
    data: np.ndarray | float,
    from_bits: int = 16,
    to_bits: int = 12,
    ccd_gain_left: float = 4.9,
    ccd_gain_right: float = 4.9,
    ccd_offset: float = 100,
    ccd_read_noise: float = 17,
    overwrite_table: bool = False,
) -> np.ndarray:
    """
    Square root decode between specified bitrate values.

    Parameters
    ----------
    data
        Input encoded data array
    from_bits
        Specified bitrate of encoded image to unpack
    to_bits
        Specified bitrate of output data (decoded)
    ccd_gain_left
        CCD gain (left side of CCD) [e / DN]
    ccd_gain_right
        CCD gain (right side of CCD) [e / DN]
    ccd_offset
        CCD bias level [DN]
    ccd_read_noise
        CCD read noise level [DN]
    overwrite_table
        Toggle to regenerate and overwrite existing decoding table

    Returns
    -------
    np.ndarray
        Square root decoded version of the input image

    """
    table_name_left = (
        TABLE_PATH
        + "tab_fb"
        + str(from_bits)
        + "_tb"
        + str(to_bits)
        + "_g"
        + str(ccd_gain_left)
        + "_b"
        + str(ccd_offset)
        + "_r"
        + str(ccd_read_noise)
        + ".npy"
    )

    table_name_right = (
        TABLE_PATH
        + "tab_fb"
        + str(from_bits)
        + "_tb"
        + str(to_bits)
        + "_g"
        + str(ccd_gain_right)
        + "_b"
        + str(ccd_offset)
        + "_r"
        + str(ccd_read_noise)
        + ".npy"
    )

    if not os.path.isfile(table_name_left) or overwrite_table:
        table_left = generate_decode_sqrt_table(from_bits, to_bits, ccd_gain_left, ccd_offset, ccd_read_noise)

        if not os.path.isdir(TABLE_PATH):
            os.makedirs(TABLE_PATH, exist_ok=True)

        np.save(table_name_left, table_left)
    else:
        table_left = np.load(table_name_left)

    if not os.path.isfile(table_name_right) or overwrite_table:
        table_right = generate_decode_sqrt_table(from_bits, to_bits, ccd_gain_right, ccd_offset, ccd_read_noise)

        if not os.path.isdir(TABLE_PATH):
            os.makedirs(TABLE_PATH, exist_ok=True)

        np.save(table_name_right, table_right)
    else:
        table_right = np.load(table_name_right)

    data = data.copy()
    data[0:data.shape[1]//2,:] = decode_sqrt_by_table(data[0:data.shape[1]//2,:], table_left)
    data[data.shape[1]//2:,:] = decode_sqrt_by_table(data[data.shape[1]//2:,:], table_right)

    return data


def encode_sqrt(data: np.ndarray | float, from_bits: int = 16, to_bits: int = 12) -> np.ndarray:
    """
    Square root encode between specified bitrate values.

    Parameters
    ----------
    data
        Input data array
    from_bits
        Specified bitrate of original input image
    to_bits
        Specified bitrate of output encoded image

    Returns
    -------
    np.ndarray
        Encoded version of input data

    """
    data = np.round(data).astype(np.int32).clip(0, None)
    factor = np.array(2 ** (2 * to_bits - from_bits))
    data_scaled_by_factor = np.round(data * factor).astype(np.int32)

    return np.floor(np.sqrt(data_scaled_by_factor)).astype(np.int32)


def decode_sqrt_simple(data: np.ndarray | float, from_bits: int = 16, to_bits: int = 12) -> np.ndarray:
    """
    Perform a simple decoding using the naive squaring strategy.

    Parameters
    ----------
    data
        Input data array
    from_bits
        Specified bitrate of original input image
    to_bits
        Specified bitrate of output encoded image

    Returns
    -------
    np.ndarray
        Decoded version of input data

    """
    data = np.round(data).astype(np.int32).clip(0, None)
    factor = 2.0 ** (2 * to_bits - from_bits)

    return np.round(np.square(data) / factor).astype(np.int32)


def noise_pdf(
    data_value: np.ndarray | float,
    ccd_gain: float = 4.9,
    ccd_offset: float = 100,
    ccd_read_noise: float = 17,
    n_sigma: int = 5,
    n_steps: int = 10000,
) -> tuple:
    """
    Generate a probability distribution function (pdf) from an input data value.

    Parameters
    ----------
    data_value
        Input data value
    ccd_gain
        CCD gain [e/DN]
    ccd_offset
        CCD bias level [DN]
    ccd_read_noise
        CCD read noise level [DN]
    n_sigma
        Number of sigma steps
    n_steps
        Number of data steps


    Returns
    -------
    np.ndarray
        Data step distribution
    normal
        Data normal distribution

    """
    # Use camera calibration to get an e-count
    electrons = np.clip((data_value - ccd_offset) * ccd_gain, 1, None)

    # Shot noise, converted back to DN
    poisson_sigma = np.sqrt(electrons) / ccd_gain

    # Total sigma is quadrature sum of fixed & shot
    sigma = np.sqrt(poisson_sigma**2 + ccd_read_noise**2)

    dn_steps = np.arange(-n_sigma * sigma, n_sigma * sigma, sigma * n_sigma * 2 / n_steps)

    # Explicitly calculate the Gaussian/normal PDF at each step
    normal = np.exp(-dn_steps * dn_steps / sigma / sigma / 2)

    # Easier to normalize numerically than to account for missing tails
    normal = normal / np.sum(normal)

    return data_value + dn_steps, normal


def mean_b_offset(
    data_value: float,
    from_bits: int = 16,
    to_bits: int = 12,
    ccd_gain: float = 4.9,
    ccd_offset: float = 100,
    ccd_read_noise: float = 17,
) -> float:
    """
    Compute an offset from the naive and robust decoding processes.

    Parameters
    ----------
    data_value
        Input data value [DN]
    from_bits
        Specified bitrate of encoded image to unpack
    to_bits
        Specified bitrate of output data (decoded)
    ccd_gain
        CCD gain [e/DN]
    ccd_offset
        CCD bias level [DN]
    ccd_read_noise
        CCD read noise level [DN]

    Returns
    -------
    float
        Generated decoding value for use in constructing a decoding table

    """
    naive_decoded_value = decode_sqrt_simple(data_value, from_bits, to_bits)

    # Generate distribution around naive value
    (values, weights) = noise_pdf(naive_decoded_value, ccd_gain, ccd_offset, ccd_read_noise)

    # Ignore values below the offset -- which break the noise model
    weights = weights * (values >= ccd_offset)

    if np.sum(weights) < 0.95:
        return 0

    weights = weights / np.sum(weights)

    # Encode the entire value distribution
    data_values = encode_sqrt(values, from_bits, to_bits)

    # Decode the entire value distribution to find the net offset
    net_offset = decode_sqrt_simple(data_values, from_bits, to_bits)

    # Expected value of the entire distribution
    expected_value = np.sum(net_offset * weights)

    # Return ΔB.
    return expected_value - naive_decoded_value


def decode_sqrt_corrected(
    data_value: float,
    from_bits: int = 16,
    to_bits: int = 12,
    ccd_gain: float = 4.9,
    ccd_offset: float = 100,
    ccd_read_noise: float = 17,
) -> float:
    """
    Compute an individual decoding value for an input data value.

    Parameters
    ----------
    data_value
        Input data value [DN]
    from_bits
        Specified bitrate of encoded image to unpack
    to_bits
        Specified bitrate of output data (decoded)
    ccd_gain
        CCD gain [e/DN]
    ccd_offset
        CCD bias level [DN]
    ccd_read_noise
        CCD read noise level [DN]

    Returns
    -------
    float
        Generated decoding value for use in constructing a decoding table

    """
    s1p = decode_sqrt_simple(data_value + 1, from_bits, to_bits)
    s1n = decode_sqrt_simple(data_value - 1, from_bits, to_bits)

    width = (s1p - s1n) / 4

    fixed_sigma = np.sqrt(ccd_read_noise**2 + width**2)

    of = mean_b_offset(data_value, from_bits, to_bits, ccd_gain, ccd_offset, fixed_sigma)

    return decode_sqrt_simple(data_value, from_bits, to_bits) - of


def generate_decode_sqrt_table(
    from_bits: int = 16,
    to_bits: int = 12,
    ccd_gain: float = 4.9,
    ccd_offset: float = 100,
    ccd_read_noise: float = 17,
) -> np.ndarray:
    """
    Generate a square root decode table between specified bitrate values and CCD parameters.

    Parameters
    ----------
    from_bits
        Specified bitrate of encoded image to unpack
    to_bits
        Specified bitrate of output data (decoded)
    ccd_gain
        CCD gain [e/DN]
    ccd_offset
        CCD bias level [DN]
    ccd_read_noise
        CCD read noise level [DN]

    Returns
    -------
    table
        Generated square root decoding table

    """
    table = np.zeros(ceil(2**to_bits))

    for i in range(table.size):
        table[i] = decode_sqrt_corrected(i, from_bits, to_bits, ccd_gain, ccd_offset, ccd_read_noise)

    return table


def decode_sqrt_by_table(data: np.ndarray | float, table: np.ndarray) -> np.ndarray:
    """
    Generate a square root decode table between specified bitrate values and CCD parameters.

    Parameters
    ----------
    data
        Input encoded data array
    table
        Square root decoding table

    Returns
    -------
    np.ndarray
        Decoded version of input data

    """
    rounded_data = np.round(data).astype(np.int32)
    if np.any(rounded_data > table.shape[0]):
        warnings.warn("Data values exceeded square-root encoding lookup table.", DataValueWarning)
    data = rounded_data.clip(0, table.shape[0]-1)

    return table[data]


@punch_task
def decode_sqrt_data(data_object: NDCube, overwrite_table: bool = False) -> NDCube:
    """
    Prefect task in the pipeline to decode square root encoded data.

    Parameters
    ----------
    data_object : NDCube
        the object you wish to decode
    overwrite_table
        Toggle to regenerate and overwrite existing decoding table

    Returns
    -------
    NDCube
        a modified version of the input with the data square root decoded

    """
    data = data_object.data

    from_bits = data_object.meta["RAWBITS"].value
    to_bits = data_object.meta["COMPBITS"].value

    ccd_gain_left = data_object.meta["GAINLEFT"].value
    ccd_gain_right = data_object.meta["GAINRGHT"].value
    ccd_offset = data_object.meta["OFFSET"].value
    ccd_read_noise = 17  # DN  # TODO: make this not a hardcoded value!

    decoded_data = decode_sqrt(
        data,
        from_bits=from_bits,
        to_bits=to_bits,
        ccd_gain_left=ccd_gain_left,
        ccd_gain_right=ccd_gain_right,
        ccd_offset=ccd_offset,
        ccd_read_noise=ccd_read_noise,
        overwrite_table=overwrite_table,
    )
    data_object.data[...] = decoded_data[...]  # TODO: we need to make sure the data type changes?

    data_object.meta.history.add_now("LEVEL0-decode-sqrt", "image square root decoded")

    return data_object



from datetime import datetime

import astropy.units as u
import cv2 as cv
import matplotlib as mpl
import matplotlib.cm as cm
import matplotlib.pyplot as plt
import numpy as np
from astropy.io import fits
from astropy.nddata import StdDevUncertainty
from astropy.wcs import WCS
from ndcube import NDCube
from sunpy.sun import constants

from punchbowl.data import load_ndcube_from_fits
from punchbowl.data.meta import NormalizedMetadata


def calc_ylims(ycen_band_rs: np.ndarray, r_band_width: float, arcsec_per_px: float) -> tuple[int, int]:
    """
    Convert y-coordinates of lower and upper row of bands to array indices for slicing.

    Parameters
    ----------
    ycen_band_rs : np.ndarray
        y-coordinates of center of band in solar radii

    r_band_width : float
        Half-width of each radial band in solar radii

    arcsec_per_px : float
        Radial pixel scale in arcsec/px in the polar-remapped images

    Returns
    -------
    list
        Lower and upper Numpy array indices of the radial band

    """
    # Unless we have a cropped image, bottom axis of the polar transform should be at 0 Rs.
    origin_rs = 0
    origin_arcsec = origin_rs * arcsec_per_px
    ycen_band_arcsec = ycen_band_rs * constants.average_angular_size.value   # center of the radial band in arcsec
    rband_width_arcsec = r_band_width * constants.average_angular_size.value  # width of the radial band in arcsec
    ylo_band_idx = ((ycen_band_arcsec - rband_width_arcsec) - origin_arcsec) / arcsec_per_px   # lower index of the band
    yhi_band_idx = ((ycen_band_arcsec + rband_width_arcsec) - origin_arcsec) / arcsec_per_px   # upper index of the band
    return [ylo_band_idx, yhi_band_idx]


def preprocess_image(data: NDCube, max_radius_px: int, num_azimuth_bins: int, az_bin: int) -> np.ndarray:
    """
    Normalize and preprocess FITS image by removing bad values and scaling.

    Parameters
    ----------
    data: NDCube
        Input data NDCube

    max_radius_px : int
        Maximum radius to include for polar remapping

    num_azimuth_bins : int
        Number of azimuthal samples to use in polar remapping

    az_bin: int
        Binning factor for binning the polar remapped image over the azimuth. The binning rule is currently numpy.mean()

    Returns
    -------
    np.ndarray, dict
        - Preprocecess polar-remapped image
        - associated metadata

    """
    # Replace with appropriate preprocessing needed to clean-up. We need to have finite values for the polar remap
    image = data.data[0,...]
    header = data.meta.to_fits_header(wcs=data.wcs)

    image[~np.isfinite(image)] = 0

    polar_image = cv.warpPolar(image.astype(np.float64), [int(max_radius_px), int(num_azimuth_bins)],
                               [header["CRPIX1"], header["CRPIX2"]], max_radius_px, cv.INTER_CUBIC)

    polar_image_binned = polar_image.T.reshape([polar_image.shape[1],
                                                polar_image.shape[0] // az_bin, az_bin]).mean(axis=2)

    # Remove background radially by taking the mean over the radial axis (From Craig's original implementation)
    polar_image_binned_radial_bkg = polar_image_binned - np.mean(polar_image_binned, axis=0)
    # Flat-fielding further: dividing by RMS value along the radial axis
    ff_rms = np.sqrt(np.mean(polar_image_binned_radial_bkg ** 2, axis=0))
    processed_image = polar_image_binned_radial_bkg / ff_rms
    # Clean-up, as divide by zero will occur
    processed_image[~np.isfinite(processed_image)] = 0

    polar_header = {
        "NAXIS": 2,
        "CTYPE1": "HPLN-CAR",
        "NAXIS1": processed_image.shape[1],
        "CDELT1": 360 / processed_image.shape[1],
        "CUNIT1": "deg",
        "CRPIX1": 0.5,
        "CRVAL1": 0,
        "CTYPE2": "HPLT-CAR",
        "NAXIS2": processed_image.shape[0],
        "CDELT2": 45 * 3600 / max_radius_px,
        "CUNIT2": "arcsec",
        "CRPIX2": processed_image.shape[0]//2 + 0.5,
        "CRVAL2": (processed_image.shape[0]//2 + 0.5) * (45 * 3600 / max_radius_px),
        "DATE-OBS": header["DATE-OBS"],
    }

    return processed_image, polar_header


def calculate_cross_correlation(image1: np.ndarray, image2: np.ndarray,
                                offsets: np.ndarray, delta_px: int,
                                central_offset: int) -> np.ndarray:
    """
    Perform cross-correlation for a range of offsets.

    Parameters
    ----------
    image1 : np.ndarray
        First image array for correlation

    image2 : np.ndarray
        Second, time-offset image array for correlation

    offsets : np.ndarray
        Array of pixel offsets to iterate over for cross-correlation

    delta_px : int
        Pixel offset increment between samples

    central_offset : int
        Central offset from which to start correlation

    Returns
    -------
    np.ndarray
        Accumulated cross-correlation array over all offsets

    """
    # Initialize accumulator array
    acc = np.zeros((len(offsets), image1.shape[0], image1.shape[1]), dtype=float)
    for jj, offset_index in enumerate(offsets):
        # The two images need to be shifted from each other at each iteration.
        # We first calculate the overall shift, then divide it in two parts, each part being used to shift each image
        this_of = int(delta_px * (offset_index - (len(offsets) - 1) / 2)) + central_offset
        # Amount of shift for image
        offset_1 = int(this_of / 2)
        offset_2 = int(this_of) - offset_1

        # Padding the images symmetrically according to the direction of the shift. The padding rule is to replicate the
        # values at the nearest edge
        if offset_1 < 0:
            padded_image1 = np.pad(image1, ((0, -offset_1), (0, 0)), mode="edge")[abs(offset_1):image1.shape[0] +
                                                                                  abs(offset_1), :]
        else:
            padded_image1 = np.pad(image1, ((offset_1, 0), (0, 0)), mode="edge")[:image1.shape[0], :]

        if offset_2 < 0:
            padded_image2 = np.pad(image2, ((-offset_2, 0), (0, 0)), mode="edge")[:image2.shape[0], :]
        else:
            padded_image2 = np.pad(image2, ((0, offset_2), (0, 0)), mode="edge")[offset_2:image2.shape[0] + offset_2, :]

        acc[jj, :, :] += padded_image1 * padded_image2

    return acc


def accumulate_cross_correlation_across_frames(files: list, delta_t: int, sparsity: int, n_ofs: int,
                                               max_radius_deg: int, num_azimuth_bins: int, az_bin: int,
                                               delta_px: int, central_offset: int) -> np.ndarray:
    """
    Accumulate cross-correlation across frames in a list of FITS files.

    Parameters
    ----------
    files : list
        List of file paths to FITS files

    delta_t : int
        Frame offset (in frames) between time-offset image pairs

    sparsity : int
        Interval between frames to skip when accumulating cross-correlation

    n_ofs : int
        Number of pixel offsets to use in cross-correlation

    max_radius_deg : int
        Maximum radius in degrees to include for polar remapping

    num_azimuth_bins : int
        Number of azimuthal samples to use in polar remapping

    az_bin: int
        Binning factor for binning the polar remapped image over the azimuth. The binning rule is currently numpy.mean()

    delta_px : int
        Pixel offset increment between samples

    central_offset : int
        Central offset from which to start correlation


    Returns
    -------
    np.ndarray
        Accumulated cross-correlation array over all frames and offsets

    """
    data = load_ndcube_from_fits(files[0])
    header = data.meta.to_fits_header(wcs=data.wcs)


    max_radius_px = max_radius_deg / header["CDELT1"]
    polar_sample, _ = preprocess_image(data, max_radius_px, num_azimuth_bins, az_bin)

    acc = np.zeros((n_ofs, polar_sample.shape[0], polar_sample.shape[1]), dtype=float)
    n = 0
    for i in range(0, len(files) - delta_t, delta_t * sparsity):
        data1 = load_ndcube_from_fits(files[i])

        data2 = load_ndcube_from_fits(files[i + delta_t])

        prepped_image1, _ = preprocess_image(data1, max_radius_px, num_azimuth_bins, az_bin)
        prepped_image2, _ = preprocess_image(data2, max_radius_px, num_azimuth_bins, az_bin)

        acc += calculate_cross_correlation(prepped_image1, prepped_image2, np.arange(n_ofs), delta_px, central_offset)

        n += 1

    acc /= n

    return acc


def compute_all_bands(acc: np.ndarray, ycen_band_rs: np.ndarray, r_band_half_width: float,
                      arcsec_per_px: float, velocity_azimuth_bins: int,
                      x_kps: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
    """
    Compute speed and sigma for all radial bands.

    Parameters
    ----------
    acc : np.ndarray
        Cross-correlation array accumulated across frames

    ycen_band_rs : np.ndarray
        y-coordinates of band centers in solar radii

    r_band_half_width : float
        Half-width of each radial band in solar radii

    arcsec_per_px : float
        Radial pixel scale in arcsec/px in the polar-remapped images

    velocity_azimuth_bins : int
        Number of azimuthal bins in the output flow maps

    x_kps : np.ndarray
        Array mapping pixel offsets to speed in km/s

    Returns
    -------
    tuple
        Tuple containing:
        - np.ndarray : Average speed per angular bin for each radial band
        - np.ndarray : Sigma (standard deviation) of speed per angular bin for each radial band

    """
    ylohi = calc_ylims(ycen_band_rs, r_band_half_width, arcsec_per_px)
    # Determine spike location (index of the correlation peak) in the cross-correlation array
    spike_location = np.where(x_kps < 0)[0].max() + 2

    avg_speeds = []
    sigmas = []
    for (ylo, yhi) in zip(*ylohi, strict=False):
        acc_k = acc[:, int(ylo):int(yhi) + 1, ...].mean(axis=1)
        # The modulus must be zero
        azimuth_bin_size = acc_k.shape[1] // velocity_azimuth_bins
        avcor_rbins_theta = acc_k.reshape(acc_k.shape[0], azimuth_bin_size, velocity_azimuth_bins)

        speedmax_idx_per_thbin = np.array(
            [avcor_rbins_theta[spike_location:, :, i].argmax(axis=0) +
             spike_location for i in range(velocity_azimuth_bins)])
        speedmax_per_theta = x_kps[speedmax_idx_per_thbin]
        avg_speeds.append(speedmax_per_theta.mean(axis=1))
        sigmas.append(speedmax_per_theta.std(axis=1) / np.sqrt(azimuth_bin_size))

    return np.array(avg_speeds), np.array(sigmas)


def process_corr(files: list, arcsec_per_px:float, expected_kps_windspeed: float, delta_t: int, sparsity: int,
                 delta_px: int, ycens: np.ndarray, r_band_half_width: float, n_ofs: int, max_radius_deg: int,
                 num_azimuth_bins: int, az_bin: int, velocity_azimuth_bins: int) -> tuple[np.ndarray, np.ndarray]:
    """
    Process the cross-correlation across frames  in a list of FITS files with associated average speeds.

    Parameters
    ----------
    files : list
        List of file paths to FITS files

    arcsec_per_px: float
        pixel scale in arcsec over the radial axis in the polar-remapped image

    expected_kps_windspeed: float
        Expected Wind Speed in km/s for narrowing the cross-correlation

    delta_t : float
        Time offset (in nb of frames) between for an image pair

    sparsity : int
        Interval between frames to skip when accumulating cross-correlation

    delta_px : int
        Pixel offset increment between samples

    ycens: np.ndarray
        y-coordinates of center of bands in solar radii

    r_band_half_width : float
        Half-width of each radial band in solar radii

    n_ofs : int
        Number of pixel offsets to use in cross-correlation

    max_radius_deg : int
        Maximum radius in degrees to include for polar remapping

    num_azimuth_bins : int
        Number of azimuthal samples to use in polar remapping

    az_bin: int
        Binning factor for binning the polar remapped image over the azimuth. The binning rule is currently numpy.mean()

    velocity_azimuth_bins : int
        Number of azimuthal bins in the output flow maps


    Returns
    -------
    [np.ndarray, np.ndarray]
        Average speed and 1-sigma uncertainty over radius and angular bins

    """
    # Expected windspeed in pixels
    time_cadence_sec = 4 * 60  # time cadence of punch images in seconds (4 minutes)
    arcsec_km = (1*u.arcsec).to(u.rad).value * 1*u.au.to(u.km)
    expected_px_windspeed =  expected_kps_windspeed / (arcsec_per_px * arcsec_km ) * time_cadence_sec
    # Central offset to start correlation from.
    central_offset = int(delta_t * expected_px_windspeed)
    # Calculate speed mapping for offsets in km/s
    x_pix = delta_px * (np.arange(n_ofs) - (n_ofs - 1) / 2) + central_offset
    x_kps = x_pix / central_offset * expected_kps_windspeed
    # Accumulate cross-correlation across frames
    acc = accumulate_cross_correlation_across_frames(files, delta_t, sparsity, n_ofs, max_radius_deg, num_azimuth_bins,
                                                     az_bin, delta_px, central_offset)
    # Compute average speeds and sigma for each radial band and latitudinal bin
    avg_speeds, sigmas = compute_all_bands(acc, ycens, r_band_half_width, arcsec_per_px, velocity_azimuth_bins, x_kps)

    return avg_speeds, sigmas


def plot_flow_map(filename: str | None, data: NDCube, cmap: str = "magma") -> None:
    """
    Plot polar maps of the radial flows.

    Parameters
    ----------
    filename: str
        Output plot filename. If None, the figure is not saved out.

    data: NDCube
        Flow tracking data NDCube

    cmap : str, optional
        Colormap for the plot (default is 'magma')

    Returns
    -------
    fig
        The generated Matplotlib Figure

    """
    speeds = data.data
    sigmas = data.uncertainty.array

    ycen_band_rs = np.fromstring(data.meta["YCENS"].value[1:-1], dtype=float, sep=",")
    rbands = np.fromstring(data.meta["RBANDS"].value[1:-1], dtype=int, sep=",")
    velocity_azimuth_bins = data.meta["PLTBINS"].value

    thetas = np.linspace(0, 2 * np.pi, velocity_azimuth_bins + 1)

    plt.close("all")
    fig = plt.figure(figsize=(20, 8))

    vmin = speeds.min()
    vmax = speeds.max()
    norm = mpl.colors.Normalize(vmin=vmin, vmax=vmax, clip=True)
    mapper = cm.ScalarMappable(norm=norm, cmap=cmap)

    for i, ridx in enumerate(rbands):

        signal = np.append(speeds[ridx], speeds[ridx][0])
        error = np.append(sigmas[ridx], sigmas[ridx][0])

        ax = fig.add_subplot(1, len(rbands), i + 1, projection="polar")
        ax.plot(thetas, signal, "k-")
        ax.fill_between(thetas, signal - error, signal + error, alpha=0.3, color="gray")

        colors = np.array([mapper.to_rgba(v) for v in signal])
        for theta, value, err, color in zip(thetas, signal, error, colors, strict=False):
            ax.plot(theta, value, "o", color=color, ms=4)
            ax.errorbar(theta, value, yerr=err, lw=2, capsize=3, color=color)

        ax.set_title(f"Altitude = {ycen_band_rs[ridx]} Rs")
        ax.set_ylim(50, 1.05 * vmax)
        ax.set_rlabel_position(270)

    cbar_ax = fig.add_axes([0.11, 0.2, 0.8, 0.03])
    plt.colorbar(mapper, cax=cbar_ax, orientation="horizontal").ax.set_xlabel("Speed (km/s)")
    if filename is not None:
        plt.savefig(filename)
    return fig


def track_velocity(files: list[str],
                   delta_t: int = 12,
                   sparsity: int = 2,
                   n_ofs: int = 151,
                   delta_px: int = 2,
                   expected_kps_windspeed: int = 300,
                   r_band_half_width: float = 0.5,
                   max_radius_deg: int = 45,
                   num_azimuth_bins: int = 1440*8,
                   az_bin: int = 4,
                   velocity_azimuth_bins: int = 36,
                   ycens: np.ndarray | None = None,
                   rbands: list[int] | None = None) -> NDCube:
    """
    Generate velocity map using flow tracking.

    Parameters
    ----------
    files : list[str]
        List of file paths for input data

    delta_t : int, optional
        Time offset in frames between images

    sparsity : int, optional
        Frame skip interval for averaging

    n_ofs : int, optional
        Number of spatial offsets for cross-correlation

    delta_px : int, optional
        Pixel offset increment per sample

    expected_kps_windspeed : int, optional
        Expected wind speed in km/s

    r_band_half_width : float, optional
        Half-width of each radial band in solar radii

    max_radius_deg : int, optional
        The maximum radius in degrees

    num_azimuth_bins : int, optional
        Number of azimuthal bins in the polar remapped images

    az_bin : int, optional
        Binning factor for binning the polar remapped image over the azimuth

    velocity_azimuth_bins : int, optional
        Number of azimuthal bins in the output flow maps

    ycens : numpy.ndarray, optional
        Radial band centers in solar radii

    rbands : list[int], optional
        Indices of radial bands to visualize

    Returns
    -------
    ndcube.NDCube
        The generated velocity map

    """
    # Set defaults for missing input parameters
    if ycens is None:
        ycens = np.arange(7, 14.5, 0.5)
    if rbands is None:
        rbands = [0, 4, 8, 14]

    files.sort()

    if len(files) < 2:
        msg = "At least to input files must be provided for flow tracking"
        raise ValueError(msg)

    # Data preprocessing
    data0 = load_ndcube_from_fits(files[0])
    header1 = data0.meta.to_fits_header(wcs=data0.wcs)
    _, polar_header1 = preprocess_image(data0, max_radius_deg/header1["CDELT1"], num_azimuth_bins, az_bin)

    avg_speeds, sigmas = process_corr(files, polar_header1["CDELT2"], expected_kps_windspeed,
                                      delta_t, sparsity, delta_px, ycens, r_band_half_width, n_ofs,
                                      max_radius_deg, num_azimuth_bins, az_bin, velocity_azimuth_bins)

    output_meta = NormalizedMetadata.load_template("VAM", "3")

    with fits.open(files[0]) as hdul:
        output_meta["DATE-BEG"] = hdul[1].header["DATE-BEG"]

    with fits.open(files[-1]) as hdul:
        output_meta["DATE-END"] = hdul[1].header["DATE-END"]

    date_beg = datetime.strptime(output_meta["DATE-BEG"].value, "%Y-%m-%dT%H:%M:%S").astimezone()
    date_end = datetime.strptime(output_meta["DATE-END"].value, "%Y-%m-%dT%H:%M:%S").astimezone()
    date_avg = (date_beg + (date_end - date_beg) / 2).strftime("%Y-%m-%dT%H:%M:%S")
    output_meta["DATE-AVG"] = date_avg
    output_meta["DATE-OBS"] = date_avg

    output_meta["DELTAT"] = delta_t
    output_meta["SPARSITY"] = sparsity
    output_meta["N_OFS"] = n_ofs
    output_meta["DELTA_PX"] = delta_px
    output_meta["KPSEXP"] = expected_kps_windspeed
    output_meta["BANDWDTH"] = r_band_half_width
    output_meta["MAXRAD"] = max_radius_deg
    output_meta["AZMBINS"] = num_azimuth_bins
    output_meta["AZMBINF"] = az_bin
    output_meta["PLTBINS"] = velocity_azimuth_bins
    output_meta["YCENS"] = np.array2string(ycens, separator=",", max_line_width=10_000)
    output_meta["RBANDS"] = np.array2string(np.array(rbands), separator=",", max_line_width=10_000)

    wcs = WCS(naxis=2)
    wcs.wcs.ctype = "radius","azimuth"
    wcs.wcs.cunit = "solRad","rad"
    wcs.wcs.cdelt = 1, 0.5
    wcs.wcs.crpix = 0, 0
    wcs.wcs.crval = 0, 0
    wcs.wcs.cname = "solar radii", "azimuth"
    wcs.array_shape = avg_speeds.shape

    return NDCube(data = avg_speeds,
                  uncertainty=StdDevUncertainty(sigmas),
                  meta = output_meta,
                  wcs = wcs)



import os
import argparse

import pandas as pd

from punchbowl.data import NormalizedMetadata
from punchbowl.data.meta import load_level_spec, load_spacecraft_def

LEVELS = ["0", "1", "2", "3", "L", "Q"]


def parse_args():
    args = argparse.ArgumentParser()
    args.add_argument("-d", "--dir", help="path to the header yaml folder", required=True)
    return args.parse_args()


def check_all_files_exist(paths):
    for level in LEVELS:
        assert f"Level{level}.yaml" in paths, f"Level{level}.yaml not found"
    assert "spacecraft.yaml" in paths, "spacecraft.yaml not found"
    assert "omniheader.csv" in paths, "omniheader.csv not found"


def validate_omniheader(directory):
    omniheader = pd.read_csv(os.path.join(directory, "omniheader.csv"), na_filter=False)
    expected_columns = {"SECTION", "TYPE", "KEYWORD", "VALUE", "COMMENT", "DATATYPE", "NULLABLE", "MUTABLE", "DEFAULT"}
    assert set(omniheader.columns) == expected_columns, "Omniheader columns do not match expected"


def construct_all_product_headers(directory, level):
    level_path = os.path.join(directory, f"Level{level}.yaml")
    level_spec = load_level_spec(level_path)
    product_keys = list(level_spec["Products"].keys())
    crafts = load_spacecraft_def().keys()
    product_keys = sorted(list(set([pc.replace("?", craft) for craft in crafts for pc in product_keys])))
    for pc in product_keys:
        try:
            meta = NormalizedMetadata.load_template(pc, level)
        except Exception as e:
            assert False, f"failed to create {pc} for level {level} because: {e}"


if __name__ == "__main__":
    # provide directory with contents
    args = parse_args()

    check_all_files_exist(set(os.listdir(args.dir)))

    validate_omniheader(args.dir)

    for level in LEVELS:
        construct_all_product_headers(args.dir, level)

    print("No problems were found in the headers!")



from datetime import UTC, datetime

import numpy as np
import pytest
from astropy.nddata import StdDevUncertainty
from astropy.wcs import WCS
from ndcube import NDCube
from prefect.logging import disable_run_logger
from pytest import fixture

from punchbowl.data import NormalizedMetadata
from punchbowl.data.units import split_ccd_array
from punchbowl.exceptions import DataValueWarning
from punchbowl.level1.sqrt import decode_sqrt, decode_sqrt_data, decode_sqrt_simple, encode_sqrt


# Some test inputs
@fixture
def sample_punchdata():
    """
    Generate a sample PUNCH data object for testing
    """

    data = np.random.random([2048, 2048])
    uncertainty = StdDevUncertainty(np.sqrt(np.abs(data)))
    wcs = WCS(naxis=2)
    wcs.wcs.ctype = "HPLN-ARC", "HPLT-ARC"
    wcs.wcs.cunit = "deg", "deg"
    wcs.wcs.cdelt = 0.01, 0.01
    wcs.wcs.crpix = 1024, 1024
    wcs.wcs.crval = 0, 0
    wcs.wcs.cname = "HPC lon", "HPC lat"
    wcs.array_shape = data.shape
    meta = NormalizedMetadata.load_template("PM1", "0")
    meta['DATE-OBS'] = str(datetime(2023, 1, 1, 0, 0, 1))

    punchdata_obj = NDCube(data=data, uncertainty=uncertainty, wcs=wcs, meta=meta)

    punchdata_obj.meta['RAWBITS'] = 16
    punchdata_obj.meta['COMPBITS'] = 10
    punchdata_obj.meta['GAINLEFT'] = 1.0/4.9
    punchdata_obj.meta['GAINRGHT'] = 1.0/4.9
    punchdata_obj.meta['OFFSET'] = 100

    return punchdata_obj


def test_encoding():
    arr_dim = 2048
    arr = np.random.random([arr_dim, arr_dim]) * (2 ** 16)

    encoded_arr = encode_sqrt(arr, from_bits=16, to_bits=10)

    assert encoded_arr.shape == arr.shape
    assert np.max(encoded_arr) <= 2 ** 10


def test_decoding():
    arr_dim = 2048
    arr = np.random.random([arr_dim, arr_dim]) * (2 ** 10)

    decoded_arr = decode_sqrt_simple(arr, from_bits=10, to_bits=16)

    assert decoded_arr.shape == arr.shape
    assert np.max(decoded_arr) <= 2 ** 16


def test_decoding_exceed_table():
    arr_dim = 2048
    arr = np.random.random([arr_dim, arr_dim]) * (2 ** 10) + 1

    with pytest.warns(DataValueWarning):
        decoded_arr = decode_sqrt(arr, from_bits=16, to_bits=10)
        assert decoded_arr.shape == arr.shape
        assert np.max(decoded_arr) <= 2 ** 16


@pytest.mark.parametrize('from_bits, to_bits', [(16, 10), (16, 11), (16, 12)])
def test_encode_then_decode(from_bits, to_bits):
    arr_dim = 2048
    ccd_gain_left = 1.0 / 4.9  # DN/electron
    ccd_gain_right = 1.0 / 4.9  # DN/electron
    ccd_offset = 100  # DN
    ccd_read_noise = 17  # DN

    original_arr = (np.random.random([arr_dim, arr_dim]) * (2 ** from_bits)).astype(int)

    encoded_arr = encode_sqrt(original_arr, from_bits, to_bits)
    decoded_arr = decode_sqrt(encoded_arr,
                              from_bits=from_bits,
                              to_bits=to_bits,
                              ccd_gain_left=ccd_gain_left,
                              ccd_gain_right=ccd_gain_right,
                              ccd_offset=ccd_offset,
                              ccd_read_noise=ccd_read_noise)

    ccd_gain = split_ccd_array(original_arr.shape, ccd_gain_left, ccd_gain_right)

    noise_tolerance = np.sqrt(original_arr / ccd_gain) * ccd_gain

    test_coords = np.where(original_arr > 150)

    assert np.all(np.abs(original_arr[test_coords] - decoded_arr[test_coords]) <= noise_tolerance[test_coords])


def test_decode_sqrt_data_task(sample_punchdata):
    """
    Test the decode_sqrt_data prefect task using a test harness
    """

    with disable_run_logger():
        output_punchdata = decode_sqrt_data.fn(sample_punchdata, overwrite_table=True)
        assert isinstance(output_punchdata, NDCube)
        assert output_punchdata.data.shape == (2048, 2048)




import astropy.units as u
import solpolpy
from ndcube import NDCollection, NDCube
from prefect import get_run_logger

from punchbowl.prefect import punch_task


def resolve_polarization(data_list: list[NDCube]) -> list[NDCube]:
    """
    Take a set of input data in the camera MZP frame and convert to the solar MZP frame.

    Parameters
    ----------
    data_list : List[NDCube]
        List of NDCube objects on which to resolve polarization

    Returns
    -------
    List[NDCube]
        modified version of the input with polarization resolved

    """
    # Unpack data into a NDCollection object
    data_dictionary = list(zip(["M", "Z", "P"], data_list, strict=False))
    input_collection = NDCollection(data_dictionary)
    data_collection = NDCollection([(k, NDCube(data=input_collection[k].data,
                                 wcs=input_collection[k].wcs,
                                 meta={"POLAR": input_collection[k].meta["POLAR"].value * u.degree,
                                       "POLAROFF": input_collection[k].meta["POLAROFF"].value * u.degree,
                                       "POLARREF": str(input_collection[k].meta["POLARREF"])}))
                       for k in ["M", "Z", "P"]])

    out = []
    resolved_data_collection = solpolpy.resolve(data_collection,
                                                "mzpsolar",
                                                imax_effect=True)
    for key in resolved_data_collection:
        resolved_data_collection[key].meta = input_collection[key].meta
        resolved_data_collection[key].uncertainty = input_collection[key].uncertainty
        out.append(resolved_data_collection[key])

    return out


@punch_task
def resolve_polarization_task(data_list: list[NDCube | None]) -> list[NDCube | None]:
    """
    Prefect task for polarization resolving.

    Parameters
    ----------
    data_list : List[NDCube]
        List of NDCube objects on which to resolve polarization

    Returns
    -------
    List[NDCube]
        modified version of the input with polarization resolved

    """
    logger = get_run_logger()

    if None in data_list:
        logger.info("Skipping polarization resolution because one of the images was None.")
        return [None, None, None]

    logger.info("resolve_polarization started")
    data_list = resolve_polarization(data_list)
    logger.info("resolve_polarization ended")

    for data_object in data_list:
        data_object.meta.history.add_now("LEVEL2-resolve_polarization", "polarization resolved")
    return data_list



from __future__ import annotations

import os
from datetime import datetime

import astropy.coordinates
import astropy.time
import astropy.units as u
import astropy.wcs.wcsapi
import numpy as np
import sunpy.coordinates
import sunpy.map
from astropy.coordinates import GCRS, EarthLocation, SkyCoord, StokesSymbol, custom_stokes_symbol_mapping, get_sun
from astropy.time import Time
from astropy.wcs import WCS
from astropy.wcs.utils import add_stokes_axis_to_wcs
from sunpy.coordinates import frames

_ROOT = os.path.abspath(os.path.dirname(__file__))
PUNCH_STOKES_MAPPING = custom_stokes_symbol_mapping({10: StokesSymbol("pB", "polarized brightness"),
                                                     11: StokesSymbol("B", "total brightness")})


def extract_crota_from_wcs(wcs: WCS) -> u.deg:
    """Extract CROTA from a WCS."""
    delta_ratio = wcs.wcs.cdelt[1] / wcs.wcs.cdelt[0]
    return np.arctan2(wcs.wcs.pc[1, 0]/delta_ratio, wcs.wcs.pc[0, 0]) * u.rad


def calculate_helio_wcs_from_celestial(wcs_celestial: WCS,
                                       date_obs: astropy.time.Time | str | None=None,
                                       data_shape: tuple[int, int] | None=None) -> tuple[WCS, float]:
    """Calculate the helio WCS from a celestial WCS."""
    if not date_obs:
        date_obs = wcs_celestial.wcs.dateobs
    if data_shape is None:
        data_shape = wcs_celestial.array_shape
    if not date_obs:
        msg = "dateobs must be provided, either as an argument or inside the WCS"
        raise ValueError(msg)
    date_obs = Time(date_obs)
    is_3d = len(data_shape) == 3

    # we're at the center of the Earth
    test_loc = EarthLocation.from_geocentric(0, 0, 0, unit=u.m)
    test_gcrs = SkyCoord(test_loc.get_gcrs(date_obs))

    # follow the SunPy tutorial from here
    # https://docs.sunpy.org/en/stable/generated/gallery/units_and_coordinates/radec_to_hpc_map.html#sphx-glr-generated-gallery-units-and-coordinates-radec-to-hpc-map-py
    reference_coord = SkyCoord(
        wcs_celestial.wcs.crval[0] * u.Unit(wcs_celestial.wcs.cunit[0]),
        wcs_celestial.wcs.crval[1] * u.Unit(wcs_celestial.wcs.cunit[1]),
        frame="gcrs",
        obstime=date_obs,
        obsgeoloc=test_gcrs.cartesian,
        obsgeovel=test_gcrs.velocity.to_cartesian(),
        distance=test_gcrs.hcrs.distance,
    )

    reference_coord_hp = reference_coord.transform_to(frames.Helioprojective(observer=test_gcrs))
    projection_code = wcs_celestial.wcs.ctype[0][-3:] if "-" in wcs_celestial.wcs.ctype[0] else ""

    hdr = sunpy.map.make_fitswcs_header(
        np.empty(data_shape),
        reference_coord_hp,
        reference_pixel=wcs_celestial.wcs.crpix * u.pix - 1 * u.pix,
        scale=np.abs(wcs_celestial.wcs.cdelt) * u.deg / u.pix,
        projection_code=projection_code)

    wcs_helio = WCS(hdr, naxis=2)
    wcs_helio.wcs.set_pv(wcs_celestial.wcs.get_pv())

    rotation_angle = -compute_hp_to_eq_rotation_angle(wcs_helio, date_obs)
    rotation_angle -= extract_crota_from_wcs(wcs_celestial)
    new_pc_matrix = calculate_pc_matrix(rotation_angle, wcs_helio.wcs.cdelt)
    wcs_helio.wcs.pc = new_pc_matrix
    wcs_helio.cpdis1 = wcs_celestial.cpdis1
    wcs_helio.cpdis2 = wcs_celestial.cpdis2

    if is_3d:
        wcs_helio = astropy.wcs.utils.add_stokes_axis_to_wcs(wcs_helio, 2)
    wcs_helio.array_shape = data_shape

    p_angle = get_p_angle(date_obs)

    return wcs_helio, p_angle


def get_sun_ra_dec(dt: datetime) -> tuple[float, float]:
    """Get the position of the Sun in right ascension and declination."""
    position = get_sun(Time(str(dt), scale="utc"))
    return position.ra.value, position.dec.value


def calculate_pc_matrix(crota: float, cdelt: tuple[float, float]) -> np.ndarray:
    """
    Calculate a PC matrix given CROTA and CDELT.

    Parameters
    ----------
    crota : float
        rotation angle from the WCS
    cdelt : float
        pixel size from the WCS

    Returns
    -------
    np.ndarray
        PC matrix

    """
    return np.array(
        [
            [np.cos(crota), -np.sin(crota) * (cdelt[0] / cdelt[1])],
            [np.sin(crota) * (cdelt[1] / cdelt[0]), np.cos(crota)],
        ],
    )


def _times_are_equal(time_1: astropy.time.Time, time_2: astropy.time.Time) -> bool:
    # Ripped from sunpy, modified
    # Checks whether times are equal
    if isinstance(time_1, astropy.time.Time) and isinstance(time_2, astropy.time.Time):
        # We explicitly perform the check in TAI to avoid possible numerical precision differences
        # between a time in UTC and the same time after a UTC->TAI->UTC conversion
        return np.all(time_1.tai == time_2.tai)

    # We also deem the times equal if one is None
    return time_1 is None or time_2 is None


def get_p_angle(time: str="now") -> u.deg:
    """
    Get the P angle.

    Return the position (P) angle for the Sun at a specified time, which is the angle between
    geocentric north and solar north as seen from Earth, measured eastward from geocentric north.
    The range of P is +/-26.3 degrees.

    Parameters
    ----------
    time : {parse_time_types}
        Time to use in a parse_time-compatible format

    Returns
    -------
    out : `~astropy.coordinates.Angle`
        The position angle

    """
    obstime = sunpy.coordinates.sun.parse_time(time)

    # Define the frame where its Z axis is aligned with geocentric north
    geocentric = astropy.coordinates.GCRS(obstime=obstime)

    return sunpy.coordinates.sun._sun_north_angle_to_z(geocentric)  # noqa: SLF001


@astropy.coordinates.frame_transform_graph.transform(
    astropy.coordinates.FunctionTransform,
    sunpy.coordinates.Helioprojective,
    astropy.coordinates.GCRS)
def hpc_to_gcrs(HPcoord, GCRSframe):  # noqa: ANN201, N803, ANN001
    """Convert helioprojective to GCRS."""
    if not _times_are_equal(HPcoord.obstime, GCRSframe.obstime):
        raise ValueError("Obstimes are not equal")  # noqa: EM101
    obstime = HPcoord.obstime or GCRSframe.obstime

    # Compute the three angles we need
    position = astropy.coordinates.get_sun(obstime)
    ra, dec = position.ra.rad, position.dec.rad
    p = -get_p_angle(obstime).rad

    # Prepare rotation matrices for each
    p_matrix = np.array([
        [1, 0, 0],
        [0, np.cos(p), -np.sin(p)],
        [0, np.sin(p), np.cos(p)],
    ])

    ra_matrix = np.array([
        [np.cos(ra), -np.sin(ra), 0],
        [np.sin(ra), np.cos(ra), 0],
        [0, 0, 1],
    ])

    dec_matrix = np.array([
        [np.cos(-dec), 0, np.sin(-dec)],
        [0, 1, 0],
        [-np.sin(-dec), 0, np.cos(-dec)],
    ])

    # Compose the matrices
    matrix = ra_matrix @ dec_matrix @ p_matrix

    # Extract the input coordinates and negate the HP latitude,
    # since it increases in the opposite direction from RA.
    if HPcoord._is_2d:  # noqa: SLF001
        rep = astropy.coordinates.UnitSphericalRepresentation(
            -HPcoord.Tx, HPcoord.Ty)
    else:
        rep = astropy.coordinates.SphericalRepresentation(
            -HPcoord.Tx, HPcoord.Ty, HPcoord.distance)

    # Apply the transformation
    rep = rep.to_cartesian()
    rep = rep.transform(matrix)

    # Match the input representation. (If the input was UnitSpherical, meaning there's no
    # distance coordinate, this drops the distance coordinate.)
    rep = rep.represent_as(type(HPcoord.data))

    # Put the computed coordinates into the output frame
    return GCRSframe.realize_frame(rep)


@astropy.coordinates.frame_transform_graph.transform(
    astropy.coordinates.FunctionTransform,
    astropy.coordinates.GCRS,
    sunpy.coordinates.Helioprojective)
def gcrs_to_hpc(GCRScoord, Helioprojective): # noqa: ANN201, N803, ANN001
    """Convert GCRS to HPC."""
    if not _times_are_equal(GCRScoord.obstime, Helioprojective.obstime):
        raise ValueError("Obstimes are not equal")  # noqa: EM101
    obstime = GCRScoord.obstime or Helioprojective.obstime

    # Compute the three angles we need
    position = astropy.coordinates.get_sun(obstime)
    ra, dec = position.ra.rad, position.dec.rad
    p = -get_p_angle(obstime).rad

    # Prepare rotation matrices for each
    p_matrix = np.array([
        [1, 0, 0],
        [0, np.cos(p), -np.sin(p)],
        [0, np.sin(p), np.cos(p)],
    ])

    ra_matrix = np.array([
        [np.cos(ra), -np.sin(ra), 0],
        [np.sin(ra), np.cos(ra), 0],
        [0, 0, 1],
    ])

    dec_matrix = np.array([
        [np.cos(-dec), 0, np.sin(-dec)],
        [0, 1, 0],
        [-np.sin(-dec), 0, np.cos(-dec)],
    ])

    # Compose the matrices
    old_matrix = ra_matrix @ dec_matrix @ p_matrix
    matrix = np.linalg.inv(old_matrix)

    # Extract the input coordinates and negate the HP latitude,
    # since it increases in the opposite direction from RA.
    if isinstance(GCRScoord.data, astropy.coordinates.UnitSphericalRepresentation):
        rep = astropy.coordinates.UnitSphericalRepresentation(
            GCRScoord.ra, GCRScoord.dec)  # , earth_distance(obstime))
    else:
        rep = astropy.coordinates.SphericalRepresentation(
            GCRScoord.ra, GCRScoord.dec, GCRScoord.distance)

    # Apply the transformation
    rep = rep.to_cartesian()
    rep = rep.transform(matrix)

    # Match the input representation. (If the input was UnitSpherical, meaning there's no
    # distance coordinate, this drops the distance coordinate.)
    rep = rep.represent_as(type(GCRScoord.data))

    if isinstance(rep, astropy.coordinates.UnitSphericalRepresentation):
        rep = astropy.coordinates.UnitSphericalRepresentation(
            -rep.lon, rep.lat)  # , earth_distance(obstime))
    else:
        rep = astropy.coordinates.SphericalRepresentation(
            -rep.lon, rep.lat, rep.distance)

    # Put the computed coordinates into the output frame
    return Helioprojective.realize_frame(rep)


def calculate_celestial_wcs_from_helio(wcs_helio: WCS,
                                       date_obs: astropy.time.Time | str | None=None,
                                       data_shape: tuple[int, int] | None=None) -> WCS:
    """Calculate the celestial WCS from a helio WCS."""
    if not date_obs:
        date_obs = wcs_helio.wcs.dateobs
    if data_shape is None:
        data_shape = wcs_helio.array_shape
    if not date_obs:
        msg = "dateobs must be provided, either as an argument or inside the WCS"
        raise ValueError(msg)
    is_3d = len(data_shape) == 3

    old_crval = SkyCoord(wcs_helio.wcs.crval[0] * u.deg, wcs_helio.wcs.crval[1] * u.deg,
                         frame="helioprojective", observer="earth", obstime=date_obs)
    new_crval = old_crval.transform_to(GCRS)

    cdelt1 = np.abs(wcs_helio.wcs.cdelt[0]) * u.deg
    cdelt2 = np.abs(wcs_helio.wcs.cdelt[1]) * u.deg

    projection_code = wcs_helio.wcs.ctype[0][-3:] if "-" in wcs_helio.wcs.ctype[0] else ""
    if projection_code:  # noqa: SIM108
        new_ctypes = (f"RA---{projection_code}", f"DEC--{projection_code}")
    else:
        new_ctypes = "RA", "DEC"

    wcs_celestial = WCS(naxis=2)
    wcs_celestial.wcs.ctype = new_ctypes
    wcs_celestial.wcs.cunit = ("deg", "deg")
    wcs_celestial.wcs.cdelt = (-cdelt1.to(u.deg).value, cdelt2.to(u.deg).value)
    wcs_celestial.wcs.crpix = (wcs_helio.wcs.crpix[0], wcs_helio.wcs.crpix[1])
    wcs_celestial.wcs.crval = (new_crval.ra.to(u.deg).value,  new_crval.dec.to(u.deg).value)
    wcs_celestial.wcs.set_pv(wcs_helio.wcs.get_pv())
    wcs_celestial.wcs.dateobs = wcs_helio.wcs.dateobs
    wcs_celestial.wcs.mjdobs = wcs_helio.wcs.mjdobs
    wcs_celestial.cpdis1 = wcs_helio.cpdis1
    wcs_celestial.cpdis2 = wcs_helio.cpdis2

    rotation_angle = compute_hp_to_eq_rotation_angle(wcs_helio, date_obs)
    rotation_angle += extract_crota_from_wcs(wcs_helio)
    new_pc_matrix = calculate_pc_matrix(rotation_angle, wcs_helio.wcs.cdelt)
    wcs_celestial.wcs.pc = new_pc_matrix

    if is_3d:
        wcs_celestial = add_stokes_axis_to_wcs(wcs_celestial, 2)
    wcs_celestial.array_shape = data_shape

    return wcs_celestial


def project_vec_onto_plane(vec1: np.ndarray, vec2: np.ndarray) -> np.ndarray:
    """Project vec1 onto the plane perpendicular to vec2."""
    # First project vec1 onto vec2
    v1_projected = np.dot(vec1, vec2) * vec2 / np.dot(vec2, vec2)
    # Now subtract that from vec1 to get the component perpendicular to vec2
    return vec1 - v1_projected


def angle_between_vectors(vec1: np.ndarray, vec2: np.ndarray, plane_normal: np.ndarray) -> np.ndarray:
    """
    Compute the angle between vec1 and vec2.

    The angle is computed in the plane normal to plane_normal. This normal vector also sets the handedness and the sign
    of the angle.
    """
    plane_normal = plane_normal / np.linalg.norm(plane_normal)
    return np.arctan2(np.dot(np.cross(vec1, vec2), plane_normal), np.dot(vec1, vec2))


def compute_hp_to_eq_rotation_angle(wcs_helio: WCS, date_obs: str | Time | None=None) -> u.Quantity:
    """
    Compute the necessary rotation angle to align a celestial WCS to a helioprojective one.

    The computed angle includes the effect of the P angle but not the rotation of the helioprojective WCS

    Parameters
    ----------
    wcs_helio : WCS
        A helioprojective WCS for which we're producing a corresponding celestial WCS. Only CRVAL is used from this WCS
    date_obs : str | Time | None
        The observation date. If not provided, it will be pulled from wcs_helio. If not provided in wcs_helio, it must
        be specified here.

    Returns
    -------
    rotation_angle : Quantity
        The rotation angle that should be added to that in wcs_celestial's PC matrix

    """
    if not date_obs:
        date_obs = wcs_helio.wcs.dateobs
    if not date_obs:
        msg = "dateobs must be provided, either as an argument or inside the WCS"
        raise ValueError(msg)
    # Get our CRVAL vector
    axis_sc = wcs_helio.celestial.pixel_to_world(wcs_helio.wcs.crpix[0] - 1, wcs_helio.wcs.crpix[1] - 1)
    # Make sure date_obs and observer are specified
    axis_sc = SkyCoord(axis_sc.data, frame="helioprojective", observer=axis_sc.observer or "earth", obstime=date_obs)
    axis = axis_sc.cartesian.xyz
    axis /= np.linalg.norm(axis)

    # Create a new vector pointing near north. Projected into the plane perpendicular to axis, this will point as
    # north as possible in that plane.
    northward = axis.copy()
    northward[2] += 20
    northward /= np.linalg.norm(northward)

    # Package up that vector and convert to our target coordinate system
    northward = SkyCoord(*northward, representation_type="cartesian", frame=axis_sc.frame)
    northward.representation_type = "spherical"
    rotated = northward.transform_to("gcrs")
    rotated = rotated.cartesian.xyz
    rotated /= np.linalg.norm(rotated)

    # And also transform the CRVAL axis
    axis_rotated = axis_sc.transform_to("gcrs")
    axis_rotated = axis_rotated.cartesian.xyz
    axis_rotated /= np.linalg.norm(axis_rotated)

    # Create a new northward vector in the target frame
    northward_in_new_frame = axis_rotated.copy()
    northward_in_new_frame[2] += 20
    northward_in_new_frame /= np.linalg.norm(northward_in_new_frame)

    # Project the old and new northward vectors into the plane perpendicular to axis
    northward_in_new_frame_proj = project_vec_onto_plane(northward_in_new_frame, axis_rotated)
    rotated_proj = project_vec_onto_plane(rotated, axis_rotated)

    return -angle_between_vectors(northward_in_new_frame_proj, rotated_proj, axis_rotated).to(u.deg)


def load_trefoil_wcs() -> tuple[astropy.wcs.WCS, tuple[int, int]]:
    """Load Level 2 trefoil world coordinate system and shape."""
    trefoil_wcs = WCS(os.path.join(_ROOT, "data", "trefoil_wcs.fits"))
    trefoil_wcs.wcs.ctype = "HPLN-ARC", "HPLT-ARC"  # TODO: figure out why this is necessary, seems like a bug
    trefoil_shape = (4096, 4096)

    trefoil_wcs.array_shape = trefoil_shape

    return trefoil_wcs, trefoil_shape


def load_quickpunch_mosaic_wcs() -> tuple[astropy.wcs.WCS, tuple[int, int]]:
    """Load Level quickPUNCH mosaic world coordinate system and shape."""
    quickpunch_mosaic_shape = (1024, 1024)
    quickpunch_mosaic_wcs = WCS(naxis=2)

    quickpunch_mosaic_wcs.wcs.crpix = quickpunch_mosaic_shape[1] / 2 + 0.5, quickpunch_mosaic_shape[0] / 2 + 0.5
    quickpunch_mosaic_wcs.wcs.crval = 0, 0
    quickpunch_mosaic_wcs.wcs.cdelt = 0.045, 0.045
    quickpunch_mosaic_wcs.wcs.ctype = "HPLN-ARC", "HPLT-ARC"

    quickpunch_mosaic_wcs.array_shape = quickpunch_mosaic_shape

    return quickpunch_mosaic_wcs, quickpunch_mosaic_shape


def load_quickpunch_nfi_wcs() -> tuple[astropy.wcs.WCS, tuple[int, int]]:
    """Load Level quickPUNCH NFI world coordinate system and shape."""
    quickpunch_nfi_shape = (1024, 1024)
    quickpunch_nfi_wcs = WCS(naxis=2)

    quickpunch_nfi_wcs.wcs.crpix = quickpunch_nfi_shape[1] / 2 + 0.5, quickpunch_nfi_shape[0] / 2 + 0.5
    quickpunch_nfi_wcs.wcs.crval = 0, 0
    quickpunch_nfi_wcs.wcs.cdelt = 30 / 3600 * 2, 30 / 3600 * 2
    quickpunch_nfi_wcs.wcs.ctype = "HPLN-TAN", "HPLT-TAN"

    quickpunch_nfi_wcs.array_shape = quickpunch_nfi_shape

    return quickpunch_nfi_wcs, quickpunch_nfi_shape



from punchbowl.data.history import History
from punchbowl.data.meta import NormalizedMetadata
from punchbowl.data.punch_io import (
    get_base_file_name,
    load_ndcube_from_fits,
    write_ndcube_to_fits,
    write_ndcube_to_quicklook,
)
from punchbowl.data.wcs import load_trefoil_wcs

__all__ = [
           "History",
           "NormalizedMetadata",
           "get_base_file_name",
           "load_ndcube_from_fits",
           "load_trefoil_wcs",
           "write_ndcube_to_fits",
    "write_ndcube_to_quicklook",
]



import copy
from collections.abc import Callable

import numpy as np
from astropy.io import fits
from astropy.wcs import WCS
from ndcube import NDCube
from thuban.pointing import refine_pointing

from punchbowl.data.wcs import calculate_celestial_wcs_from_helio, calculate_helio_wcs_from_celestial
from punchbowl.prefect import punch_task


@punch_task
def align_task(data_object: NDCube, distortion_path: str | None, mask: Callable | None = None,
               pointing_shift: tuple[float, float] = (0, 0)) -> NDCube:
    """
    Determine the pointing of the image and updates the metadata appropriately.

    Parameters
    ----------
    data_object : NDCube
        data object to align
    mask : Callable | None
        function accepting coordinates and returning them only if they are not masked out
    distortion_path: str | None
        path to a distortion model
    pointing_shift: tuple[float, float]
        offset to pre-apply to the pointing to account for boresight to satellite-x shift

    Returns
    -------
    NDCube
        a modified version of the input with the WCS more accurately determined

    """
    celestial_input = calculate_celestial_wcs_from_helio(copy.deepcopy(data_object.wcs),
                                                         data_object.meta.astropy_time,
                                                         data_object.data.shape)
    refining_data = data_object.data.copy()
    refining_data[np.isinf(refining_data)] = 0
    refining_data[np.isnan(refining_data)] = 0

    if distortion_path:
        with fits.open(distortion_path) as distortion_hdul:
            distortion_wcs = WCS(distortion_hdul[0].header, distortion_hdul)
        celestial_input.cpdis1 = distortion_wcs.cpdis1
        celestial_input.cpdis2 = distortion_wcs.cpdis2
        celestial_input.wcs.set_pv([(2, 1, distortion_wcs.wcs.get_pv()[0][-1])])
        celestial_input.wcs.cdelt = distortion_wcs.wcs.cdelt

    celestial_input.wcs.crval = (celestial_input.wcs.crval[0] - pointing_shift[0] * celestial_input.wcs.cdelt[0],
                                 celestial_input.wcs.crval[1] - pointing_shift[1] * celestial_input.wcs.cdelt[1])

    celestial_output, observed_coords, result, trial_num = refine_pointing(refining_data, celestial_input,
                                                                        num_stars=10,
                                                                        detection_threshold=5,
                                                                        background_height=16,
                                                                        background_width=16,
                                                                        max_trials=5,
                                                                        dimmest_magnitude=6.5,
                                                                        chisqr_threshold=0.1,
                                                                        edge=100,
                                                                        mask=mask,
                                                                        max_error=150,
                                                                        method="least_squares")

    recovered_wcs, _ = calculate_helio_wcs_from_celestial(celestial_output,
                                                       data_object.meta.astropy_time,
                                                       data_object.data.shape)

    if distortion_path:
        with fits.open(distortion_path) as distortion_hdul:
            distortion_wcs = WCS(distortion_hdul[0].header, distortion_hdul)
        recovered_wcs.cpdis1 = distortion_wcs.cpdis1
        recovered_wcs.cpdis2 = distortion_wcs.cpdis2

    output = NDCube(data=data_object.data,
                    wcs=recovered_wcs,
                    uncertainty=data_object.uncertainty,
                    unit=data_object.unit,
                    meta=data_object.meta)
    output.meta.history.add_now("LEVEL1-Align", "alignment done")
    return output



import os
import pathlib

import pytest
from ndcube import NDCube

from punchbowl.data.punch_io import write_ndcube_to_fits
from punchbowl.data.tests.test_punch_io import sample_ndcube
from punchbowl.data.wcs import load_trefoil_wcs
from punchbowl.level2.flow import ORDER, level2_core_flow, level2_ctm_flow

THIS_DIRECTORY = pathlib.Path(__file__).parent.resolve()


def test_core_flow_runs_with_filenames(sample_ndcube, tmpdir):
    data_list = [sample_ndcube(shape=(10, 10), code=code, level="1") for code in ORDER]
    paths = []
    for i, cube in enumerate(data_list):
        path = os.path.join(tmpdir, f"test_input_{i}.fits")
        write_ndcube_to_fits(cube, path)
        paths.append(path)
    voters = [[] for _ in data_list]

    trefoil_wcs, _ = load_trefoil_wcs()
    output = level2_core_flow(paths, voters, trefoil_wcs=trefoil_wcs, trefoil_shape=(4096, 4096))
    assert isinstance(output[0], NDCube)


def test_ctm_flow_runs_with_filenames(sample_ndcube, tmpdir):
    data_list = [sample_ndcube(shape=(10, 10), code=code, level="1") for code in ["CR1", "CR2", "CR3", "CR4"]]
    paths = []
    for i, cube in enumerate(data_list):
        path = os.path.join(tmpdir, f"test_input_{i}.fits")
        write_ndcube_to_fits(cube, path)
        paths.append(path)
    voters = [[] for _ in data_list]

    trefoil_wcs, _ = load_trefoil_wcs()
    output = level2_ctm_flow(paths, voters, trefoil_wcs=trefoil_wcs, trefoil_shape=(4096, 4096))
    assert isinstance(output[0], NDCube)

@pytest.mark.parametrize("drop_indices", [[], [1], [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11]])
def test_core_flow_runs_with_objects_and_calibration_files(sample_ndcube, drop_indices):
    data_list = [sample_ndcube(shape=(10, 10), code=code, level="1")
                 for i, code in enumerate(ORDER) if i not in drop_indices]

    trefoil_wcs, _ = load_trefoil_wcs()
    voters = [[] for _ in data_list]
    output = level2_core_flow(data_list, voters, trefoil_wcs=trefoil_wcs, trefoil_shape=(4096, 4096))
    assert isinstance(output[0], NDCube)



from datetime import UTC, datetime

import numpy as np
from astropy.nddata import StdDevUncertainty
from astropy.wcs import WCS
from ndcube import NDCube
from prefect import get_run_logger

from punchbowl.data import get_base_file_name, load_trefoil_wcs
from punchbowl.data.meta import NormalizedMetadata, set_spacecraft_location_to_earth
from punchbowl.level2.bright_structure import identify_bright_structures_task
from punchbowl.level2.merge import merge_many_clear_task, merge_many_polarized_task
from punchbowl.level2.polarization import resolve_polarization_task
from punchbowl.level2.resample import reproject_many_flow
from punchbowl.prefect import punch_flow
from punchbowl.util import average_datetime, load_image_task, output_image_task

ORDER = ["PM1", "PZ1", "PP1",
         "PM2", "PZ2", "PP2",
         "PM3", "PZ3", "PP3",
         "PM4", "PZ4", "PP4"]


@punch_flow
def level2_core_flow(data_list: list[str] | list[NDCube],
                     voter_filenames: list[list[str]],
                     trefoil_wcs: WCS | None = None,
                     trefoil_shape: tuple[int, int] | None = None,
                     output_filename: str | None = None) -> list[NDCube]:
    """Level 2 core flow."""
    logger = get_run_logger()
    logger.info("beginning level 2 core flow")

    data_list = [load_image_task(d) if isinstance(d, str) else d for d in data_list]

    if data_list:
        # order the data list so it can be processed properly
        output_dateobs = average_datetime([d.meta.datetime for d in data_list]).isoformat()
        output_datebeg = min([d.meta.datetime for d in data_list]).isoformat()
        output_dateend = max([d.meta.datetime for d in data_list]).isoformat()

        ordered_data_list: list[NDCube | None] = [None for _ in range(len(ORDER))]
        ordered_voters: list[list[str]] = [[] for _ in range(len(ORDER))]
        for i, order_element in enumerate(ORDER):
            for j, data_element in enumerate(data_list):
                typecode = data_element.meta["TYPECODE"].value
                obscode = data_element.meta["OBSCODE"].value
                if typecode == order_element[:2] and obscode == order_element[2]:
                    ordered_data_list[i] = data_element
                    ordered_voters[i] = voter_filenames[j]
        logger.info("Ordered files are "
                    f"{[get_base_file_name(cube) if cube is not None else None for cube in ordered_data_list]}")

        if trefoil_wcs is None or trefoil_shape is None:
            trefoil_wcs, trefoil_shape = load_trefoil_wcs()

        data_list = [resolve_polarization_task.submit(ordered_data_list[i:i+3]) for i in range(0, len(ORDER), 3)]
        data_list = [entry.result() for entry in data_list]
        data_list = reproject_many_flow([j for i in data_list for j in i], trefoil_wcs, trefoil_shape)
        data_list = [identify_bright_structures_task(cube, this_voter_filenames)
                     for cube, this_voter_filenames in zip(data_list, ordered_voters, strict=True)]
        output_data = merge_many_polarized_task(data_list, trefoil_wcs)
    else:
        output_dateobs = datetime.now(UTC).isoformat()
        output_datebeg = output_dateobs
        output_dateend = output_datebeg

        output_data = NDCube(
        data=np.zeros(trefoil_shape),
        uncertainty=StdDevUncertainty(np.zeros(trefoil_shape)),
        wcs=trefoil_wcs,
        meta=NormalizedMetadata.load_template("PTM", "2"),
    )

    output_data.meta["DATE"] = datetime.now(UTC).isoformat()
    output_data.meta["DATE-AVG"] = output_dateobs
    output_data.meta["DATE-OBS"] = output_dateobs
    output_data.meta["DATE-BEG"] = output_datebeg
    output_data.meta["DATE-END"] = output_dateend
    output_data = set_spacecraft_location_to_earth(output_data)

    if output_filename is not None:
        output_image_task(output_data, output_filename)

    logger.info("ending level 2 core flow")
    return [output_data]


@punch_flow
def level2_ctm_flow(data_list: list[str] | list[NDCube],
                    voter_filenames: list[list[str]],
                    trefoil_wcs: WCS | None = None,
                    trefoil_shape: tuple[int, int] | None = None,
                    output_filename: str | None = None) -> list[NDCube]:
    """Level 2 CTM flow."""
    logger = get_run_logger()
    logger.info("beginning level 2 CTM flow")

    data_list = [load_image_task(d) if isinstance(d, str) else d for d in data_list]

    if data_list:
        output_dateobs = average_datetime([d.meta.datetime for d in data_list]).isoformat()
        output_datebeg = min([d.meta.datetime for d in data_list]).isoformat()
        output_dateend = max([d.meta.datetime for d in data_list]).isoformat()

        if trefoil_wcs is None or trefoil_shape is None:
            trefoil_wcs, trefoil_shape = load_trefoil_wcs()

        data_list = reproject_many_flow(data_list, trefoil_wcs, trefoil_shape)
        data_list = [identify_bright_structures_task(cube, this_voter_filenames)
                     for cube, this_voter_filenames in zip(data_list, voter_filenames, strict=True)]
        output_data = merge_many_clear_task(data_list, trefoil_wcs)
        output_data.meta["FILEVRSN"] = data_list[0].meta["FILEVRSN"].value
    else:
        output_dateobs = datetime.now(UTC).isoformat()
        output_datebeg = output_dateobs
        output_dateend = output_datebeg

        output_data = NDCube(
        data=np.zeros(trefoil_shape),
        uncertainty=StdDevUncertainty(np.zeros(trefoil_shape)),
        wcs=trefoil_wcs,
        meta=NormalizedMetadata.load_template("CTM", "2"),
    )

    output_data.meta["DATE"] = datetime.now(UTC).isoformat()
    output_data.meta["DATE-AVG"] = output_dateobs
    output_data.meta["DATE-OBS"] = output_dateobs
    output_data.meta["DATE-BEG"] = output_datebeg
    output_data.meta["DATE-END"] = output_dateend
    output_data = set_spacecraft_location_to_earth(output_data)

    if output_filename is not None:
        output_image_task(output_data, output_filename)

    logger.info("ending level 2 CTM flow")
    return [output_data]



import os
import pathlib
from collections.abc import Callable, Generator

import astropy.units as u
import numpy as np
from ndcube import NDCube
from prefect import get_run_logger
from regularizepsf import ArrayPSFBuilder, ArrayPSFTransform, simple_functional_psf
from regularizepsf.util import calculate_covering

from punchbowl.data import NormalizedMetadata
from punchbowl.data.units import calculate_image_pixel_area, dn_to_msb
from punchbowl.level1.alignment import align_task
from punchbowl.level1.deficient_pixel import remove_deficient_pixels_task
from punchbowl.level1.despike import despike_task
from punchbowl.level1.destreak import destreak_task
from punchbowl.level1.initial_uncertainty import update_initial_uncertainty_task
from punchbowl.level1.psf import correct_psf_task
from punchbowl.level1.quartic_fit import perform_quartic_fit_task
from punchbowl.level1.sqrt import decode_sqrt_data
from punchbowl.level1.stray_light import remove_stray_light_task
from punchbowl.level1.vignette import correct_vignetting_task
from punchbowl.prefect import punch_flow
from punchbowl.util import load_image_task, output_image_task


@punch_flow
def generate_psf_model_core_flow(input_filepaths: list[str],
                                 masks: list[pathlib.Path | str] | np.ndarray | Generator = None,
                                 alpha: float = 2.0,
                                 epsilon: float = 0.3,
                                 image_shape: tuple[int, int] = (2048, 2048),
                                 psf_size: int = 32,
                                 target_fwhm: float = 3.25) -> ArrayPSFTransform:
    """Generate PSF model."""
    # Define the target PSF as a symmetric Gaussian
    center = psf_size / 2
    sigma = target_fwhm / 2.355
    @simple_functional_psf
    def target(row, col, x0=center, y0=center, sigma_x=sigma, sigma_y=sigma) -> np.ndarray:  # noqa: ANN001
        return np.exp(-(np.square(row - x0) / (2 * np.square(sigma_x))
                        + np.square(col - y0) / (2 * np.square(sigma_y))))

    image_psf, counts = ArrayPSFBuilder(psf_size).build(input_filepaths, hdu_choice=1, star_masks=masks)
    coords = calculate_covering(image_shape, psf_size)
    return ArrayPSFTransform.construct(image_psf, target.as_array_psf(coords, psf_size), alpha, epsilon)


@punch_flow
def level1_core_flow(
    input_data: list[str] | list[NDCube],
    gain_left: float = 4.9,
    gain_right: float = 4.9,
    bias_level: float = 100,
    dark_level: float = 55.81,
    read_noise_level: float = 17,
    bitrate_signal: int = 16,
    quartic_coefficient_path: str | pathlib.Path | Callable | None = None,
    despike_sigclip: float = 50,
    despike_sigfrac: float = 0.25,
    despike_objlim: float = 160.0,
    despike_niter: int = 10,
    despike_cleantype: str = "meanmask",
    exposure_time: float = 49 * 1000,
    readout_line_time: float = 163/2148,
    reset_line_time: float = 163/2148,
    vignetting_function_path: str | Callable | None = None,
    stray_light_path: str | None = None,
    deficient_pixel_map_path: str | None = None,
    deficient_pixel_method: str = "median",
    deficient_pixel_required_good_count: int = 3,
    deficient_pixel_max_window_size: int = 10,
    psf_model_path: str | Callable | None = None,
    distortion_path: str | None = None,
    output_filename: list[str] | None = None,
    max_workers: int | None = None,
    mask_path: str | None = None,
    pointing_shift: tuple[float, float] = (0, 0),
) -> list[NDCube]:
    """Core flow for level 1."""
    logger = get_run_logger()

    logger.info("beginning level 1 core flow")

    output_data = []
    for i, this_data in enumerate(input_data):
        data = load_image_task(this_data) if isinstance(this_data, str) else this_data

        if data.meta["ISSQRT"].value:
            data = decode_sqrt_data(data)
        data = update_initial_uncertainty_task(data,
                                               bias_level=bias_level,
                                               dark_level=dark_level,
                                               gain_left=gain_left,
                                               gain_right=gain_right,
                                               read_noise_level=read_noise_level,
                                               bitrate_signal=bitrate_signal,
                                               )
        data = perform_quartic_fit_task(data, quartic_coefficient_path)

        if data.meta["OBSCODE"].value == "4":
            scaling = {"gain_left": data.meta["GAINLEFT"].value * u.photon / u.DN,
                       "gain_right": data.meta["GAINRGHT"].value * u.photon / u.DN,
                       "wavelength": 530. * u.nm,
                       "exposure": data.meta["EXPTIME"].value * u.s,
                       "aperture": 49.57 * u.mm ** 2}
        else:
            scaling = {"gain_left": data.meta["GAINLEFT"].value * u.photon / u.DN,
                       "gain_right": data.meta["GAINRGHT"].value * u.photon / u.DN,
                       "wavelength": 530. * u.nm,
                       "exposure": data.meta["EXPTIME"].value * u.s,
                       "aperture": 34 * u.mm ** 2}
        pixel_scale = calculate_image_pixel_area(data.wcs, data.data.shape).to(u.sr) / u.pixel
        scaling["pixel_scale"] = pixel_scale
        data.data[:, :] = np.clip(dn_to_msb(data.data[:, :], data.wcs, **scaling), a_min=0, a_max=None)
        data.uncertainty.array[:, :] = dn_to_msb(data.uncertainty.array[:, :], data.wcs, **scaling)

        data = despike_task(data,
                            despike_sigclip,
                            despike_sigfrac,
                            despike_objlim,
                            despike_niter,
                            gain_left,  # TODO: despiking should handle the gain more completely
                            read_noise_level,
                            despike_cleantype)
        data = destreak_task(data,
                             exposure_time=exposure_time,
                             reset_line_time=reset_line_time,
                             readout_line_time=readout_line_time,
                             max_workers=max_workers)
        data = correct_vignetting_task(data, vignetting_function_path)
        data = remove_deficient_pixels_task(data,
                                            deficient_pixel_map_path,
                                            required_good_count=deficient_pixel_required_good_count,
                                            max_window_size=deficient_pixel_max_window_size,
                                            method=deficient_pixel_method)
        data = remove_stray_light_task(data, stray_light_path)
        data = correct_psf_task(data, psf_model_path, max_workers=max_workers)

        observatory = int(data.meta["OBSCODE"].value)
        if observatory < 4:
            alignment_mask = lambda x, y: (x > 100) * (x < 1900) * (y > 250) * (y < 1900)
        else:
            alignment_mask = lambda x, y: (((x < 824) + (x > 1224)) * ((y < 824) + (y > 1224))
                                           * (x > 100) * (x < 1900) * (y > 100) * (y < 1900))
        data = align_task(data, distortion_path, mask=alignment_mask, pointing_shift=pointing_shift)

        if mask_path:
            with open(mask_path, "rb") as f:
                b = f.read()
            mask = np.unpackbits(np.frombuffer(b, dtype=np.uint8)).reshape(2048, 2048)
            data.data *= mask
            data.uncertainty.array[mask==0] = np.inf

        # Repackage data with proper metadata
        product_code = data.meta["TYPECODE"].value + data.meta["OBSCODE"].value
        new_meta = NormalizedMetadata.load_template(product_code, "1")
        # copy over the existing values
        for key in data.meta:
            if key in new_meta:
                new_meta[key] = data.meta[key].value
        new_meta.history = data.meta.history
        new_meta["DATE-OBS"] = data.meta["DATE-OBS"].value  # TODO: do this better and fill rest of meta
        new_meta["CALPSF"] = os.path.basename(psf_model_path) if psf_model_path else ""
        new_meta["CALVI"] = os.path.basename(vignetting_function_path) if vignetting_function_path else ""
        new_meta["CALSL"] = os.path.basename(stray_light_path) if stray_light_path else ""
        new_meta["CALCF"] = os.path.basename(quartic_coefficient_path) if quartic_coefficient_path else ""
        new_meta["LEVEL"] = "1"

        data = NDCube(data=data.data, meta=new_meta, wcs=data.wcs, unit=data.unit, uncertainty=data.uncertainty)

        if output_filename is not None and i < len(output_filename) and output_filename[i] is not None:
            output_image_task(data, output_filename[i])
        output_data.append(data)
        logger.info("ending level 1 core flow")
    return output_data


@punch_flow
def levelh_core_flow(
    input_data: list[str] | list[NDCube],
    gain_left: float = 4.9,
    gain_right: float = 4.9,
    bias_level: float = 100,
    dark_level: float = 55.81,
    read_noise_level: float = 17,
    bitrate_signal: int = 16,
    psf_model_path: str | None = None,
    output_filename: str | None = None,
) -> list[NDCube]:
    """Core flow for level 0.5 also known as level H."""
    logger = get_run_logger()

    logger.info("beginning level H core flow")

    output_data = []
    for i, this_data in enumerate(input_data):
        data = load_image_task(this_data) if isinstance(this_data, str) else this_data
        data = decode_sqrt_data(data)
        data = update_initial_uncertainty_task(data,
                                               bias_level=bias_level,
                                               dark_level=dark_level,
                                               gain_left=gain_left,
                                               gain_right=gain_right,
                                               read_noise_level=read_noise_level,
                                               bitrate_signal=bitrate_signal,
                                               )

        data = correct_psf_task(data, psf_model_path)

        observatory = int(data.meta["OBSCODE"].value)
        if observatory < 4:
            def alignment_mask(x:float, y:float) -> float:
                return (x > 100) * (x < 1900) * (y > 250) * (y < 1900)
        else:
            def alignment_mask(x:float, y:float) -> float:
                return ((x < 824) + (x > 1224)) * ((y < 824) + (y > 1224)) * \
                    (x > 100) * (x < 1900) * (y > 100) * (y < 1900)
        data = align_task(data, mask=alignment_mask)

        # Repackage data with proper metadata
        product_code = data.meta["TYPECODE"].value + data.meta["OBSCODE"].value
        new_meta = NormalizedMetadata.load_template(product_code, "H")
        new_meta["DATE-OBS"] = data.meta["DATE-OBS"].value

        output_header = new_meta.to_fits_header(data.wcs)
        for key in output_header:
            if (key in data.meta) and output_header[key] == "" and (key != "COMMENT") and (key != "HISTORY"):
                new_meta[key].value = data.meta[key].value

        data = NDCube(data=data.data, meta=new_meta, wcs=data.wcs, unit=data.unit, uncertainty=data.uncertainty)

        if output_filename is not None and i < len(output_filename) and output_filename[i] is not None:
            output_image_task(data, output_filename[i])
        output_data.append(data)
        logger.info("ending level H core flow")
    return output_data



import numpy as np
import pytest
from pytest import fixture

from punchbowl.level1.quartic_fit import (
    create_coefficient_image,
    create_constant_quartic_coefficients,
    photometric_calibration,
)


@fixture
def fake_one_image(shape=(10, 10)):
    """
    Parameters
    ----------
    shape : tuple
        the shape of the image to create

    Returns
    -------
    A fake image of the given shape with values of all 1

    """
    return np.ones(shape)


@fixture
def quartic_coefficients_image():
    """
    Returns
    -------
        Quartic coefficient image with flat_coefficients of [0, 1, 2, 3, 4]
    """
    flat_coeffs = np.array([0, 1, 2, 3, 4])
    image_shape = (10, 10)
    return create_coefficient_image(flat_coeffs, image_shape)


def test_create_constant_coefficient_image():
    image_shape = (10, 20)
    coeffs = create_constant_quartic_coefficients(image_shape)
    assert coeffs.shape[1:] == image_shape
    assert coeffs.shape[0] == 5
    for i in [0, 1, 2, 4]:
        assert np.all(coeffs[i, :, :] == 0)
    assert np.all(coeffs[3, :, :] == 1)


def test_create_coeff_image_with_only_one_coeff():
    coeffs_image = create_coefficient_image(np.array([5]), (10, 10))
    assert coeffs_image.shape == (
        1,
        10,
        10,
    ), "Failed to make a coefficients image with only one coefficient"


def test_create_coeff_image_with_two_coeff():
    coeffs_image = create_coefficient_image(np.array([5, 6]), (10, 10))
    assert coeffs_image.shape == (
        2,
        10,
        10,
    ), "Failed to make a coefficients image with two coefficients"


def test_create_coefficient_image(quartic_coefficients_image):
    assert quartic_coefficients_image.shape == (
        5,
        10,
        10,
    ), "Quartic coefficient image shape is wrong."
    for i in range(quartic_coefficients_image.shape[0]):
        assert np.all(
            quartic_coefficients_image[i, :, :] == i
        ), "Quartic coefficient image coefficients were not constructed properly."


@pytest.mark.parametrize("constant", [1, 2, 3])
def test_calibration_with_single_constant_coeff(
    fake_one_image, constant, shape=(10, 10)
):
    coeffs_image = create_coefficient_image(np.array([constant]), shape)
    calibrated = photometric_calibration(fake_one_image, coeffs_image)
    assert np.all(
        calibrated == constant
    ), "Photometric calibration with a single constant failed"


def test_calibration_with_all_one_coefficients(fake_one_image, shape=(10, 10)):
    coeffs_image = create_coefficient_image(np.array([1, 1, 1, 1, 1]), shape)
    calibrated = photometric_calibration(fake_one_image, coeffs_image)
    assert np.all(
        calibrated == 5
    ), "Photometric calibration with all one coefficients failed"


def test_calibration_with_range_coefficients(fake_one_image, shape=(10, 10)):
    coeffs_image = create_coefficient_image(np.array([1, 2, 3, 4, 5]), shape)
    calibrated = photometric_calibration(fake_one_image, coeffs_image)
    assert np.all(
        calibrated == 15
    ), "Photometric calibration with range coefficients failed"



"""Creates test data with the appropriate metadata for punchbowl"""
import os
from io import BytesIO
from datetime import UTC, datetime, timedelta

import astropy.nddata
import numpy as np
import requests
from astropy.io import fits
from astropy.nddata import StdDevUncertainty
from astropy.wcs import WCS
from ndcube import NDCube

from punchbowl.data import NormalizedMetadata, get_base_file_name, write_ndcube_to_fits
from punchbowl.level1.quartic_fit import create_constant_quartic_coefficients


def create_L2_PTM_test_file(path="../punchbowl/level3/tests/data/"):
    src_file = "https://data.boulder.swri.edu/svankooten/PUNCH_L2_PTM_20241106032023_v1.fits"
    r = requests.get(src_file)
    with fits.open(BytesIO(r.content)) as hdul:
        for hdu in hdul[1:]:
            hdu.data = astropy.nddata.block_reduce(hdu.data, (1, 8, 8), np.nanmean)
            for key in ' ', 'A':
                wcs = WCS(hdu.header, key=key)
                w = wcs[:, ::8, ::8]
                hdu.header.update(w.to_header(key=key))
        file_path = os.path.join(path, "downsampled_L2_PTM.fits")
        hdul.writeto(file_path, overwrite=True)


def create_f_corona_test_data(path="../punchbowl/level3/tests/data/"):
    meta = NormalizedMetadata.load_template("CFM", "3")
    meta["DATE-OBS"] = str(datetime.now(UTC))
    wcs = WCS(naxis=2)
    for i in range(10):
        data = np.ones((3, 10, 10)) * i
        obj = NDCube(data=data, wcs=wcs, meta=meta)
        file_path = os.path.join(path, f"test_{i}.fits")
        write_ndcube_to_fits(obj, file_path, overwrite=True)


# def create_punchdata_test_data(path="../punchbowl/tests/"):
#     meta = NormalizedMetadata.load_template("CFM", "3")
#     meta["DATE-OBS"] = str(datetime.now(UTC))
#     wcs = WCS(naxis=2)
#     data = np.ones((10, 10))
#     obj = NDCube(data=data, wcs=wcs, meta=meta)
#     file_path = os.path.join(path, "test_data.fits")
#     write_ndcube_to_fits(obj, file_path, overwrite=True)


# def create_header_validation_test_data(path="../punchbowl/tests/"):
#     wcs = WCS(naxis=2)
#     m = NormalizedMetadata.load_template("PTM", "3")
#     m['DATE-OBS'] = datetime.utcnow().isoformat()
#     m['DATE-BEG'] = datetime.utcnow().isoformat()
#     m['DATE-AVG'] = (datetime.utcnow() + timedelta(minutes=2)).isoformat()
#     m['DATE-END'] = (datetime.utcnow() + timedelta(minutes=4)).isoformat()
#     m['DATE'] = (datetime.utcnow() + timedelta(hours=1)).isoformat()
#     h = m.to_fits_header(wcs=wcs)
#
#     data = np.ones((2, 4096, 4096), dtype=np.float32)
#     uncertainty = StdDevUncertainty(np.sqrt(np.abs(data)).astype(np.float32))
#
#     d = NDCube(data=data, uncertainty=uncertainty, wcs=WCS(h), meta=m)
#     file_path = os.path.join(path, "test_header_validation.fits")
#     write_ndcube_to_fits(d, file_path, overwrite=True)


def create_quartic_coefficients_test_data(path="../punchbowl/level1/tests/data/"):
    meta = NormalizedMetadata.load_template("FQ1", "1")
    meta['DATE-OBS'] = str(datetime.now(UTC))
    wcs = WCS(naxis=3)
    data = create_constant_quartic_coefficients((10, 10))
    obj = NDCube(data=data, wcs=wcs, meta=meta)
    file_path = os.path.join(path, "test_quartic_coeffs.fits")
    write_ndcube_to_fits(obj, file_path, overwrite=True)


def create_vignetting_test_data(path="../punchbowl/level1/tests/data/"):
    meta = NormalizedMetadata.load_template("GR1", "1")
    meta['DATE-OBS'] = str(datetime(2024,2, 22, 16,34, 25))
    wcs = WCS(naxis=2)
    data = np.random.random((10, 10))
    obj = NDCube(data=data, wcs=wcs, meta=meta)
    file_path = os.path.join(path, get_base_file_name(obj) + '.fits')
    write_ndcube_to_fits(obj, file_path, overwrite=True)

def create_stray_light_test_data(path="../punchbowl/level1/tests/data/"):
    meta = NormalizedMetadata.load_template("SM1", "1")
    meta['DATE-OBS'] = str(datetime(2024,2, 22, 16,34, 25))
    wcs = WCS(naxis=2)
    data = np.random.random((10, 10))
    obj = NDCube(data=data, wcs=wcs, meta=meta)
    file_path = os.path.join(path, get_base_file_name(obj) + '.fits')
    write_ndcube_to_fits(obj, file_path, overwrite=True)

if __name__ == "__main__":
    # create_header_validation_test_data()
    create_f_corona_test_data()
    # create_punchdata_test_data()
    create_quartic_coefficients_test_data()
    create_stray_light_test_data()
    create_vignetting_test_data()



from collections.abc import Callable

import numpy as np
import scipy.signal
from ndcube import NDCube
from prefect import get_run_logger
from sklearn.decomposition import PCA
from threadpoolctl import threadpool_limits

from punchbowl.prefect import punch_task
from punchbowl.util import load_image_task


@punch_task()
def pca_filter(input_cube: NDCube, files_to_fit: list[NDCube | Callable | str],
               n_components: int=50, med_filt: int=5) -> None:
    """Run PCA-based filtering."""
    logger = get_run_logger()
    all_files_to_fit = np.empty((len(files_to_fit), *input_cube.data.shape), dtype=input_cube.data.dtype)
    for i, input_file in enumerate(files_to_fit):
        if isinstance(input_file, NDCube):
            all_files_to_fit[i] = input_file.data
        elif isinstance(input_file, str):
            all_files_to_fit[i] = load_image_task(input_file, include_provenance=False, include_uncertainty=False).data
        elif isinstance(input_file, Callable):
            input_file(all_files_to_fit[i])
        else:
            raise TypeError(f"Invalid type {type(input_file)} for input file")
    logger.info(
        f"Loaded {len(all_files_to_fit)} images to fit; filling {all_files_to_fit.nbytes/1024**3:.2f} GB")

    with threadpool_limits(30):
        pca = PCA(n_components=n_components)
        pca.fit(all_files_to_fit.reshape((len(all_files_to_fit), -1)))
        logger.info("Fitting finished")

        transformed = pca.transform(input_cube.data.reshape((1, -1)))

        if med_filt:
            for i in range(len(pca.components_)):
                comp = pca.components_[i].reshape(input_cube.data.shape)
                comp = scipy.signal.medfilt2d(comp, med_filt)
                pca.components_[i] = comp.ravel()
            logger.info("Median smoothing finished")

        reconstructed = pca.inverse_transform(transformed).reshape(input_cube.data.shape)
        input_cube.data -= reconstructed



from __future__ import annotations

import os
import typing as t
from datetime import UTC, datetime
from collections import OrderedDict
from collections.abc import Mapping

import astropy.units as u
import numpy as np
import pandas as pd
import yaml
from astropy.coordinates import GCRS, SkyCoord
from astropy.io import fits
from astropy.io.fits import Header
from astropy.time import Time
from astropy.wcs import WCS
from dateutil.parser import parse as parse_datetime
from ndcube import NDCube
from sunpy.coordinates import frames, get_earth, sun
from sunpy.coordinates.sun import _sun_north_angle_to_z
from sunpy.map import solar_angular_radius

from punchbowl.data.history import History
from punchbowl.data.wcs import calculate_celestial_wcs_from_helio, extract_crota_from_wcs, get_p_angle
from punchbowl.exceptions import MissingMetadataError

ValueType = int | str | float
_ROOT = os.path.abspath(os.path.dirname(__file__))
REQUIRED_HEADER_KEYWORDS = ["SIMPLE", "BITPIX", "NAXIS", "EXTEND"]
DISTORTION_KEYWORDS = ["CPDIS1", "CPDIS2", "DP1", "DP2"]
WCS_OMITTED_KEYWORDS = ["TIMESYS", "DATE-OBS", "DATE-BEG", "DATE-AVG", "DATE-END", "TELAPSE",
                        "RSUN_REF", "DSUN_OBS", "CRLN_OBS", "CRLT_OBS", "HGLN_OBS", "HGLT_OBS"]


def load_omniheader(path: str | None = None) -> pd.DataFrame:
    """Load full metadata specifications."""
    if path is None:
        path = os.path.join(_ROOT, "data", "omniheader.csv")
    return pd.read_csv(path, na_filter=False)


def load_level_spec(path: str) -> dict[str, t.Any]:
    """Load data product metadata specifications."""
    with open(path) as f:
        return yaml.safe_load(f)


def load_spacecraft_def(path: str | None = None) -> dict[str, t.Any]:
    """
    Load spacecraft metadata specifications.

    If path is None, then it loads a default from the package.
    """
    if path is None:
        path = os.path.join(_ROOT, "data", "spacecraft.yaml")
    with open(path) as f:
        return yaml.safe_load(f)


def construct_all_product_codes(level: str) -> list[str]:
    """Generate all data product codes."""
    level_path = os.path.join(_ROOT, "data", f"Level{level}.yaml")
    level_spec = load_level_spec(level_path)
    product_keys = list(level_spec["Products"].keys())
    if level in ["0", "H", "1", "2"]:
        crafts = {"1": "", "2": "", "3": "", "4": ""}.keys()
    if level in ["2", "3", "Q", "L"]:
        crafts = {"M": "", "N": ""}.keys()
    if level in ["Q", "L"]:
        crafts = {"M": "", "N": ""}.keys()

    return sorted({pc.replace("?", craft) for craft in crafts for pc in product_keys})


class MetaField:
    """The MetaField object describes a single field within the NormalizedMetadata object."""

    def __init__(
        self,
        keyword: str,
        comment: str,
        value: ValueType | None,
        datatype: t.Any,
        nullable: bool,
        mutable: bool,
        default: ValueType | None,
    ) -> None:
        """
        Create a MetaField.

        Parameters
        ----------
        keyword: str
            FITS keyword for this field
        comment: str
            FITS compliant comment for this field
        value : int, str, or float
            the value associated with this field
        datatype : int, str, or float type
            what type of data is expected for the value and default
        nullable : bool
            if true, the default will be used in the case of None for the value
        mutable : bool
            if false, the value can never be changed after creation
        default : int, str, or float
            the default value to use if value is None and nullable is True

        """
        if value is not None and not isinstance(value, datatype):
            msg = f"MetaField value and kind must match. Found kind={datatype} and value={type(value)}."
            raise TypeError(msg)
        if default is not None and not isinstance(default, datatype):
            msg = f"MetaField default and kind must match. Found kind={datatype} and default={type(default)}."
            raise TypeError(
                msg,
            )
        if len(keyword) > 8:
            msg = "Keywords must be 8 characters or shorter to comply with FITS"
            raise ValueError(msg)
        self._keyword = keyword
        self._comment = comment
        self._value = value
        self._datatype = datatype
        self.nullable = nullable
        self._mutable = mutable
        self._default = default

    @property
    def datatype(self) -> t.Any:
        """Get the data type."""
        return self._datatype

    @property
    def keyword(self) -> str:
        """Returns MetaField keyword."""
        return self._keyword

    @property
    def comment(self) -> str:
        """Returns MetaField comment."""
        return self._comment

    @property
    def value(self) -> ValueType:
        """Returns MetaField value if set, otherwise get the default. Raise an error if no default is available."""
        return self._value if self._value is not None else self.default

    @value.setter
    def value(self, value: ValueType) -> None:
        """Set value within MetaField object."""
        if not self._mutable:
            msg = "Cannot mutate this value because it is set to immutable."
            raise RuntimeError(msg)
        if self._type_matches(value, self._datatype) or value is None:
            self._value = value
        else:
            msg = f"Value of {self.keyword} was {type(value)} but must be {self._datatype}."
            raise TypeError(msg)
        self._value = value

    @property
    def default(self) -> ValueType:
        """Get the default value."""
        return self._default

    @default.setter
    def default(self, default: ValueType) -> None:
        if self._type_matches(default, self._datatype) or default is None:
            self._default = default
        else:
            msg = f"Value was {type(default)} but must be {self._datatype}."
            raise TypeError(msg)

    @staticmethod
    def _type_matches(value: ValueType, field_type: t.Any) -> bool:
        numpy_equivalents = {int: np.integer, float: np.floating}
        if (isinstance(value, field_type) or
                (field_type in numpy_equivalents and isinstance(value, numpy_equivalents[field_type]))):
            return True
        return field_type is float and isinstance(value, int)

    def __eq__(self, other: MetaField) -> bool:
        """Check equality."""
        if not isinstance(other, MetaField):
            msg = f"MetaFields can only be compared to their own type, found {type(other)}."
            raise TypeError(msg)
        return (
            self._keyword == other._keyword
            and self._comment == other._comment
            and self._value == other._value
            and self._datatype == other._datatype
            and self.nullable == other.nullable
            and self._mutable == other._mutable
            and self._default == other._default
        )

    def __int__(self) -> int:
        """Get an int."""
        if self.datatype is int:
            return int(self._value)
        raise TypeError(f"Cannot convert {self._datatype} to int.")

    def __float__(self) -> float:
        """Get a float."""
        if self.datatype is float or self.datatype is int:
            return float(self._value)
        raise TypeError(f"Cannot convert {self._datatype} to float.")

    def __str__(self) -> str:
        """Get a string."""
        return str(self._value)


class NormalizedMetadata(Mapping):
    """
    Represent Metadata consistently.

    The NormalizedMetadata object standardizes metadata and metadata access in the PUNCH pipeline. It does so by
    uniting the history and header fields while providing helpful accessors for commonly used formats of the metadata.

    Internally, the keys are always stored as FITS compliant upper-case strings. These are stored in sections.
    So the contents dictionary should have a key of a section title mapped to a dictionary of field keys mapped to
    MetaFields.
    """

    def __len__(self) -> int:
        """Return number of entry cards in NormalizedMetadata object."""
        return sum([len(section) for section in self._contents.values()])

    def __init__(
            self,
            contents: t.OrderedDict[str, t.OrderedDict[str, MetaField]],
            history: History | None = None,
            provenance: list[str] | None = None,
            wcs_section_name: str = "World Coordinate System",
    ) -> None:
        """
        Create a Normalized Metadata. Also see `from_template` as that is often more helpful.

        Parameters
        ----------
        contents: OrderedDict[str, OrderedDict[str, MetaField]]
            contents of the meta information
        history: History
            history contents for this meta field
        provenance: list[str]
            list of files used in the generation of this product
        wcs_section_name: str
            the section title for the WCS section to specially fill

        """
        self._contents = contents
        self._history = history if history is not None else History()
        self._provenance = provenance if provenance is not None else []
        self._wcs_section_name = wcs_section_name

    def keys(self) -> t.Iterable[str]:
        """Return FITS keys for collection."""
        return self.fits_keys

    def __iter__(self) -> t.Iterator[t.Any]:
        """Iterate."""
        return self._contents.__iter__()

    def __eq__(self, other: NormalizedMetadata) -> bool:
        """Check equality."""
        if not isinstance(other, NormalizedMetadata):
            msg = f"Can only check equality between two NormalizedMetadata, found {type(other)}."
            raise TypeError(msg)
        return self._contents == other._contents and self._history == other._history

    def to_fits_header(self, wcs: WCS | None = None, write_celestial_wcs: bool = True) -> Header:  # noqa: C901
        """
        Convert a constructed NormalizedMetdata object to an Astropy FITS compliant header object.

        Returns
        -------
        Header
            Astropy FITS compliant header object

        """
        hdr = fits.Header()
        for section in self._contents:
            if section != "FITS Required":
                hdr.append(
                    ("COMMENT", ("----- " + section + " ").ljust(72, "-")),
                    end=True,
                )
            # for normal sections
            for field in self._contents[section].values():
                if field.value is not None:
                    value = field.value
                elif field.value is None and field.nullable:
                    value = field.default
                else:
                    msg = f"Value is null for {field.keyword} and no default is allowed."
                    raise RuntimeError(msg)
                hdr.append(
                    (
                        field.keyword,
                        value,
                        field.comment,
                    ),
                    end=True,
                )

            # add the special WCS section
            if section == self._wcs_section_name:
                if wcs is None:
                    msg = "WCS was expected but not provided."
                    raise MissingMetadataError(msg)
                if write_celestial_wcs:
                    wcses = {"": wcs, "A": calculate_celestial_wcs_from_helio(wcs, self.astropy_time, self.shape)}
                else:
                    wcses = {"": wcs}
                if self.product_level == "0":
                    hdr.insert("CROTA", ("COMMENT","Level 0 WCS approximated from spacecraft-reported state"))
                for key, this_wcs in wcses.items():
                    if this_wcs.has_distortion:
                        wcs_header = this_wcs.to_fits()[0].header
                        for required_key in REQUIRED_HEADER_KEYWORDS:
                            del wcs_header[required_key]
                    else:
                        wcs_header = this_wcs.to_header()
                    for card in wcs_header.cards:
                        if ((key == "" and card[0] not in WCS_OMITTED_KEYWORDS)
                                or (key != "" and card[0][-1].isnumeric() and
                                    card[0] not in DISTORTION_KEYWORDS and
                                    card[0] not in WCS_OMITTED_KEYWORDS)):
                            hdr.append(
                                (
                                card[0] + key,
                                card[1],
                                card[2],
                                ),
                                end=True,
                            )
                hdr["CROTA"] = (extract_crota_from_wcs(wcs)).to(u.deg).value

        # add the history section
        for entry in self.history:
            hdr["HISTORY"] = f"{entry.datetime: %Y-%m-%dT%H:%M:%S} => {entry.source} => {entry.comment}|"

        # fill in dynamic values
        if wcs is not None:
            geocentric = GCRS(obstime=self.astropy_time)
            p_angle = _sun_north_angle_to_z(geocentric)
            center_helio_coord = SkyCoord(
                wcs.wcs.crval[0] * u.deg,
                wcs.wcs.crval[1] * u.deg,
                frame=frames.Helioprojective,
                obstime=self.astropy_time,
                observer="earth",
            )
            hdr["RSUN_ARC"] = solar_angular_radius(center_helio_coord).value
            hdr["SOLAR_EP"] = p_angle.value
            hdr["CAR_ROT"] = float(sun.carrington_rotation_number(t=self.astropy_time))

        return hdr

    def delete_section(self, section_name: str) -> None:
        """
        Delete a section of NormalizedMetadata.

        Parameters
        ----------
        section_name : str
            the section to delete

        Returns
        -------
        None

        """
        if section_name in self._contents:
            del self._contents[section_name]
        else:
            msg = f"Section {section_name} was not found."
            raise MissingMetadataError(msg)


    @classmethod
    def from_fits_header(cls, h: Header) -> NormalizedMetadata:
        """
        Construct a normalized Metadata from a PUNCH FITS header.

        Parameters
        ----------
        h : Header
            a PUNCH FITS header from Astropy

        Returns
        -------
        NormalizedMetadata
            the corresponding NormalizedMetadata

        """
        if "TYPECODE" not in h:
            msg = "TYPECODE must a field of the header"
            raise MissingMetadataError(msg)
        if "OBSCODE" not in h:
            msg = "OBSCODE must be a field of the header"
            raise MissingMetadataError(msg)
        if "LEVEL" not in h:
            msg = "LEVEL must be a field of the header"
            raise MissingMetadataError(msg)

        type_code, obs_code, level = h["TYPECODE"], h["OBSCODE"], h["LEVEL"]

        m = NormalizedMetadata.load_template(type_code + obs_code, level)

        for k, v in h.items():
            if k not in ("COMMENT", "HISTORY", "") and k in m:
                m[k] = v
        m.history = History.from_fits_header(h)

        return m

    @staticmethod
    def _match_product_code_in_level_spec(product_code: str, level_spec: dict) -> dict:
        """
        Parse the specified product code and level specification to find a corresponding set.

        Parameters
        ----------
        product_code
            Specified data product code
        level_spec
            Data product level specifications, loaded from `load_level_spec`

        Returns
        -------
        Dict
            Product code specification parsed from file

        """
        if product_code in level_spec["Products"]:
            return level_spec["Products"][product_code]
        else:  # noqa: RET505, okay structure
            type_code = product_code[:-1]
            found_type_codes = {pc[:-1] for pc in level_spec["Products"]}
            if type_code in found_type_codes:
                return level_spec["Products"][type_code + "?"]
            else:  # noqa: RET505, okay structure
                msg = f"Product code {product_code} not found in level_spec"
                raise RuntimeError(msg)

    @staticmethod
    def _load_template_files(
        omniheader_path: str, level: str, level_spec_path: str, spacecraft: str, spacecraft_def_path: str,
    ) -> tuple[dict, dict, dict]:
        """
        Load template files from specified locations.

        Parameters
        ----------
        omniheader_path
            Path to full omniheader specifications
        level
            Specified data product level
        level_spec_path
            Path to data product level specifications
        spacecraft
            Specified spacecraft code
        spacecraft_def_path
            Path to spacecraft specifications

        Returns
        -------
        Tuple
            Header specification entries

        """
        omniheader = load_omniheader(omniheader_path)
        spacecraft_def = load_spacecraft_def(spacecraft_def_path)
        if spacecraft not in spacecraft_def:
            msg = f"Spacecraft {spacecraft} not in spacecraft_def."
            raise RuntimeError(msg)

        if level is not None and level_spec_path is not None:
            msg = "Only specify the level or level_spec_path, not both."
            raise RuntimeError(msg)
        elif level is not None:  # noqa: RET506, fine structure
            level_spec_path = os.path.join(_ROOT, "data", f"Level{level}.yaml")
            level_spec = load_level_spec(level_spec_path)
        elif level_spec_path is not None:
            level_spec = load_level_spec(level_spec_path)
        else:
            msg = "Either level or level_spec_path must be defined. Found None for both."
            raise RuntimeError(msg)
        return omniheader, level_spec, spacecraft_def

    @staticmethod
    def _determine_omits_and_overrides(  # noqa: C901
        level_spec: dict,  # , not too complex
        product_def: dict,
    ) -> tuple[list[str], dict[str, str]]:
        """
        Read level specifications and product definitions and determines keywords to omit or overwrite.

        Parameters
        ----------
        level_spec
            Data product level specifications
        product_def
            Data product specifications

        Returns
        -------
        Tuple
            Keywords and values to omit and override

        """
        this_kinds = product_def["kinds"]
        omits, overrides = [], {}
        for section in level_spec["Level"]:
            if level_spec["Level"][section] is not None:
                if "omits" in level_spec["Level"][section]:
                    omits += level_spec["Level"][section]["omits"]
                if "overrides" in level_spec["Level"][section]:
                    for key, value in level_spec["Level"][section]["overrides"].items():
                        overrides[key] = value

        for kind in this_kinds:
            if kind not in level_spec["Kinds"]:
                msg = f"{kind} not found in level_spec."
                raise RuntimeError(msg)
            if "omits" in level_spec["Kinds"][kind]:
                omits += level_spec["Kinds"][kind]["omits"]
            if "overrides" in level_spec["Kinds"][kind]:
                for key, value in level_spec["Kinds"][kind]["overrides"].items():
                    overrides[key] = value

        if "omits" in product_def:
            omits += product_def["omits"]

        if "overrides" in product_def:
            for key, value in product_def["overrides"].items():
                overrides[key] = value

        return omits, overrides

    @classmethod
    def load_template(  # noqa: C901
        cls,
        product_code: str,
        level: str | None = None,
        level_spec_path: str | None = None,
        omniheader_path: str | None = None,
        spacecraft_def_path: str | None = None,
    ) -> NormalizedMetadata:
        """
        Given data product specification, loads relevant template files and constructs a NormalizedMetadata object.

        Parameters
        ----------
        product_code
            Specified data product code, a three character code like PM1
        level
            Specified data product level
        level_spec_path
            Path to data product level specifications
        omniheader_path
            Path to full omniheader specifications
        spacecraft_def_path
            Path to spacecraft specifications

        Returns
        -------
        NormalizedMetadata
            Constructed NormalizedMetadata object from template specifications

        """
        # load all needed files
        spacecraft = product_code[-1]
        omniheader, level_spec, spacecraft_def = NormalizedMetadata._load_template_files(
            omniheader_path, level, level_spec_path, spacecraft, spacecraft_def_path,
        )

        product_def = NormalizedMetadata._match_product_code_in_level_spec(product_code, level_spec)
        omits, overrides = NormalizedMetadata._determine_omits_and_overrides(level_spec, product_def)

        # construct the items to fill
        contents, history = OrderedDict(), History()

        # figure out the sections
        section_rows = np.where(omniheader["TYPE"] == "section")[0]
        section_titles = omniheader["VALUE"].iloc[section_rows]
        section_ids = omniheader["SECTION"].iloc[section_rows]

        # parse each section
        dtypes = {"str": str, "int": int, "float": float}
        for section_id, section_title in zip(section_ids, section_titles, strict=False):
            if section_title in level_spec["Level"] and section_title not in omits:
                contents[section_title] = OrderedDict()
                for i in np.where(omniheader["SECTION"] == section_id)[0][1:]:
                    e = omniheader.iloc[i]
                    if e["KEYWORD"] not in omits:
                        datatype = dtypes[e["DATATYPE"]]
                        value, default = e["VALUE"], e["DEFAULT"]
                        if e["KEYWORD"] in overrides:
                            value = overrides[e["KEYWORD"]]
                        try:
                            if datatype is str:
                                value = datatype(value)
                                value = value.format(**spacecraft_def[spacecraft])
                            elif (datatype is int) or (datatype is float):
                                value = datatype(value) if value != "" else None
                        except ValueError as err:
                            msg = f"Value was of the wrong type to parse for {e['KEYWORD']}"
                            raise RuntimeError(msg) from err

                        try:
                            if datatype is str:
                                default = datatype(default)
                                default = default.format(**spacecraft_def[spacecraft])
                            elif (datatype is int) or (datatype is float):
                                default = datatype(default) if default != "" else None
                        except ValueError as err:
                            msg = f"Default was of the wrong type to parse for {e['KEYWORD']}"
                            raise RuntimeError(msg) from err

                        contents[section_title][e["KEYWORD"]] = MetaField(
                            e["KEYWORD"],
                            e["COMMENT"].format(**spacecraft_def[spacecraft]),
                            value,
                            datatype,
                            e["NULLABLE"],
                            e["MUTABLE"],
                            default,
                        )
        return cls(contents, history)

    @property
    def sections(self) -> list[str]:
        """Returns header sections."""
        return list(self._contents.keys())

    @property
    def fits_keys(self) -> list[str]:
        """Returns fits keys in header template."""

        def flatten(xss: list) -> list:
            return [x for xs in xss for x in xs]

        return flatten([list(self._contents[section_name].keys()) for section_name in self._contents])

    @property
    def history(self) -> History:
        """Returns header history."""
        return self._history

    @history.setter
    def history(self, history: History) -> None:
        self._history = history

    @property
    def provenance(self) -> list[str]:
        """Returns file provenance."""
        return self._provenance

    @provenance.setter
    def provenance(self, provenance: list[str]) -> None:
        """Set file provenance."""
        self._provenance = provenance

    @staticmethod
    def _validate_key_is_str(key: str) -> None:
        """
        Validate that the provided key is a valid header keyword string.

        Parameters
        ----------
        key
            Header key string

        Returns
        -------
        None

        """
        if not isinstance(key, str):
            msg = f"Keys for NormalizedMetadata must be strings. You provided {type(key)}."
            raise TypeError(msg)
        if len(key) > 8:
            msg = f"Keys must be <= 8 characters long, received {key}"
            raise ValueError(msg)

    def __setitem__(self, key: str, value: t.Any) -> None:
        """
        Set specified pair of keyword and value in the NormalizedMetadata object.

        Parameters
        ----------
        key
            Header key string
        value
            Header value

        Returns
        -------
        None

        """
        for section_name, section in self._contents.items():
            if key in section:
                self._contents[section_name][key.upper()].value = value
                return

        # reaching here means we haven't returned
        msg = f"MetaField with key={key} not found."
        raise RuntimeError(msg)

    def get(self, key: str | tuple[str, int], default: t.Any | None = None) -> t.Any:
        """Get a value given a key or use a default value."""
        try:
            out = self[key]
        except KeyError:
            out = default
        return out

    def __getitem__(self, key: str | tuple[str, int]) -> t.Any:
        """
        Get specified keyword from NormalizedMetadata object.

        Parameters
        ----------
        key : str | tuple
            Header key string

        Returns
        -------
        t.Any
            Returned header value

        """
        if isinstance(key, tuple):
            key, i = key
        self._validate_key_is_str(key)

        for section_name, section in self._contents.items():
            if key in section:
                return self._contents[section_name][key.upper()]

        # reaching here means we haven't returned
        msg = f"MetaField with key={key} not found."
        raise RuntimeError(msg)

    def __delitem__(self, key: str) -> None:
        """
        Delete specified keyword entry from the NormalizedMetadata object.

        Parameters
        ----------
        key
            Header key string

        Returns
        -------
        None

        """
        self._validate_key_is_str(key)
        for section_name, section in self._contents.items():
            if key in section:
                del self._contents[section_name][key.upper()]
                return

        # reaching here means we haven't returned
        msg = f"MetaField with key={key} not found."
        raise RuntimeError(msg)

    def __contains__(self, key: str) -> bool:
        """
        Determine if the specified keyword is contained within the NormalizedMetadata object.

        Parameters
        ----------
        key
            Header key string

        Returns
        -------
        Boolean
            Value indicating if the specified keyword is contained within the NormalizedMetadata object

        """
        return any(key in section for section in self._contents.values())

    @property
    def product_level(self) -> int:
        """Returns data product level if indicated in metadata."""
        if "LEVEL" not in self:
            msg = "LEVEL is missing from the metadata."
            raise MissingMetadataError(msg)
        return self["LEVEL"].value

    @property
    def datetime(self) -> datetime:
        """Returns a datetime representation of the 'DATE-OBS' header keyword if indicated in metadata."""
        if "DATE-OBS" not in self:
            msg = "DATE-OBS is missing from the metadata."
            raise MissingMetadataError(msg)
        return parse_datetime(self["DATE-OBS"].value).replace(tzinfo=UTC)

    @property
    def shape(self) -> tuple:
        """Get the data shape in array order."""
        return tuple([self[f"NAXIS{i}"].value for i in range(self["NAXIS"].value, 0, -1)])

    @property
    def astropy_time(self) -> Time:
        """Get the date-obs as an astropy Time object."""
        return Time(self.datetime)

    @property
    def product_code(self) -> str:
        """"Retrieve the product code."""
        if "OBSCODE" not in self:
            msg = "OBSCODE is missing from the metadata."
            raise MissingMetadataError(msg)
        if "TYPECODE" not in self:
            msg = "TYPECODE is missing from the metadata."
            raise MissingMetadataError(msg)
        return self["TYPECODE"].value + self["OBSCODE"].value


def set_spacecraft_location_to_earth(input_data: NDCube) -> NDCube:
    """Update the spacecraft location metadata."""
    time_obs = input_data.meta.astropy_time

    input_data.meta["GEOD_LAT"] = 0.
    input_data.meta["GEOD_LON"] = 0.
    input_data.meta["GEOD_ALT"] = 0.

    coord = get_earth(time_obs)
    coord.observer = "earth"

    # S/C Heliographic Stonyhurst
    input_data.meta["HGLN_OBS"] = coord.heliographic_stonyhurst.lon.value
    input_data.meta["HGLT_OBS"] = coord.heliographic_stonyhurst.lat.value

    # S/C Heliographic Carrington
    input_data.meta["CRLN_OBS"] = coord.heliographic_carrington.lon.value
    input_data.meta["CRLT_OBS"] = coord.heliographic_carrington.lat.value

    input_data.meta["DSUN_OBS"] = sun.earth_distance(time_obs).to(u.m).value

    # S/C Heliocentric Earth Ecliptic
    input_data.meta["HEEX_OBS"] = coord.heliocentricearthecliptic.cartesian.x.to(u.m).value
    input_data.meta["HEEY_OBS"] = coord.heliocentricearthecliptic.cartesian.y.to(u.m).value
    input_data.meta["HEEZ_OBS"] = coord.heliocentricearthecliptic.cartesian.z.to(u.m).value

    # S/C Heliocentric Inertial
    input_data.meta["HCIX_OBS"] = coord.heliocentricinertial.cartesian.x.to(u.m).value
    input_data.meta["HCIY_OBS"] = coord.heliocentricinertial.cartesian.y.to(u.m).value
    input_data.meta["HCIZ_OBS"] = coord.heliocentricinertial.cartesian.z.to(u.m).value

    # S/C Heliocentric Earth Equatorial
    input_data.meta["HEQX_OBS"] = (coord.heliographic_stonyhurst.cartesian.x.value * u.AU).to(u.m).value
    input_data.meta["HEQY_OBS"] = (coord.heliographic_stonyhurst.cartesian.y.value * u.AU).to(u.m).value
    input_data.meta["HEQZ_OBS"] = (coord.heliographic_stonyhurst.cartesian.z.value * u.AU).to(u.m).value

    input_data.meta["SOLAR_EP"] = get_p_angle(time_obs).to(u.deg).value
    input_data.meta["CAR_ROT"] = float(sun.carrington_rotation_number(time_obs))

    return input_data
